# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import is_dataclass
from inspect import getfullargspec
from typing import TYPE_CHECKING

from .types import T, MISSING

if TYPE_CHECKING:
    from ..client import Client
    from typing import Dict, Callable, Any, Optional


def construct_client_dict(client: Client, data: Dict[...]):
    return {**data, "_client": client, "_http": client.http}


def convert(
    value: Any,
    factory: Callable[[Any], T],
    check: Optional[T] = None,
    client: Optional[Client] = None,
) -> T:
    def handle_factory() -> T:
        def fin_fac(v: Any):
            if check is not None and isinstance(v, check):
                return v

            try:
                if client and "_client" in getfullargspec(factory).args:
                    return factory(construct_client_dict(client, v))
            except TypeError:  # Building type/has no signature
                pass

            # The import has been placed locally to avoid circular imports
            # TODO: Find a way to remove this monstrosity
            from ..utils import APIObject

            if isinstance(v, APIObject):
                return v

            return factory(v)

        return (
            list(map(fin_fac, value))
            if isinstance(value, list)
            else fin_fac(value)
        )

    return MISSING if value is MISSING else handle_factory()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from typing import Optional, List

    from ...utils.snowflake import Snowflake


@dataclass
class VoiceServerUpdateEvent(APIObject):
    """Sent when a guild's voice server is updated.
    This is sent when initially connecting to voice,
    and when the current voice instance fails over to a new server.

    Attributes
    ----------
    token: :class:`str`
        Voice connection token
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        The guild this voice server update is for
    endpoint: Optional[:class:`str`]
        The voice server host
    """

    token: str
    guild_id: Snowflake
    endpoint: Optional[str] = None


@dataclass
class VoiceChannelSelectEvent(APIObject):
    """
    Sent when the client joins a voice channel

    Attributes
    ----------
    channel_id : Optional[:class:`Snowflake`]
        id of channel

    guild_id : Optional[:class:`Snowflake`]
        id of guild
    """

    channel_id: Optional[Snowflake] = None
    guild_id: Optional[Snowflake] = None


class VoiceConnectionStates(Enum):
    """
    Attributes
    ----------
    DISCONNECTED : :class:`str`
        TCP disconnected

    AWAITING_ENDPOINT : :class:`str`
        Waiting for voice endpoint

    AUTHENTICATING : :class:`str`
        TCP authenticating

    CONNECTING : :class:`str`
        TCP connecting

    CONNECTED : :class:`str`
        TCP connected

    VOICE_DISCONNECTED : :class:`str`
        TCP connected, Voice disconnected

    VOICE_CONNECTING : :class:`str`
        TCP connected, Voice connecting

    VOICE_CONNECTED : :class:`str`
        TCP connected, Voice connected

    NO_ROUTE : :class:`str`
        No route to host

    ICE_CHECKING : :class:`str`
        WebRTC ice checking

    """

    DISCONNECTED = auto()
    AWAITING_ENDPOINT = auto()
    AUTHENTICATING = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    VOICE_DISCONNECTED = auto()
    VOICE_CONNECTING = auto()
    VOICE_CONNECTED = auto()
    NO_ROUTE = auto()
    ICE_CHECKING = auto()


@dataclass
class VoiceConnectionStatusEvent(APIObject):
    """
    Sent when the client's voice connection status changes

    state : :class:`VoiceConnectionStates`
        one of the voice connection states listed below

    hostname : :class:`str`
        hostname of the connected voice server

    pings : List[:class:`int`]
        last 20 pings (in ms)

    average_ping : :class:`int`
        average ping (in ms)

    last_ping : :class:`int`
        last ping (in ms)

    """

    state: VoiceConnectionStates
    hostname: str
    pings: List[int]
    average_ping: int
    last_ping: int


@dataclass
class SpeakingStartEvent(APIObject):
    """
    Sent when a user in a subscribed voice channel speaks

    Attributes
    ----------
    user_id : :class:`Snowflake`
        id of user who started speaking
    """

    user_id: Snowflake


@dataclass
class SpeakingStopEvent(APIObject):
    """
    Sent when a user in a subscribed voice channel stops speaking

    Attributes
    ----------
    user_id : :class:`Snowflake`
        id of user who stopped speaking
    """

    user_id: Snowflake



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from time import time


class SlidingWindow:
    def __init__(self, capacity: int, time_unit: float):
        self.capacity: int = capacity
        self.time_unit: float = time_unit

        self.__cur_time: float = time()
        self.__pre_count: int = capacity
        self.__cur_count: int = 0

    def allow(self) -> bool:
        # Reset rate limit:
        if (time() - self.__cur_time) > self.time_unit:
            self.__cur_time = time()
            self.__pre_count = self.__cur_count
            self.__cur_count = 0

        # Calculate the estimated count
        passed_time = time() - self.__cur_time
        time_cnt = (self.time_unit - passed_time) / self.time_unit
        est_cnt = self.__pre_count * time_cnt + self.__cur_count

        # Request has passed the capacity
        if est_cnt > self.capacity:
            return False

        self.__cur_count += 1
        return True



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from typing import Optional, List

    from ...utils.snowflake import Snowflake


@dataclass
class WelcomeScreenChannel(APIObject):
    """Represents a welcome screen channel. This is a channel which gets
    shown on the welcome screen.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The channel's id
    description: :class:`str`
        The description shown for the channel
    emoji_id: Optional[:class:`int`]
        The emoji id, if the emoji is custom
    emoji_name: Optional[:class:`str`]
        The emoji name if custom, the unicode character if standard,
        or null if no emoji is set
    """
    channel_id: Snowflake
    description: str

    emoji_id: Optional[int] = None
    emoji_name: Optional[str] = None


@dataclass
class WelcomeScreen(APIObject):
    """Representation of a Discord guild/server welcome screen.

    Attributes
    ----------
    description: LIst[:class:`~pincer.objects.guild.welcome_screen.WelcomeScreenChannel`]
        The server description shown in the welcome screen
    welcome_channels: Optional[:class:`str`]
        The channels shown in the welcome screen, up to 5
    """  # noqa: E501
    welcome_channels: List[WelcomeScreenChannel]

    description: Optional[str] = None



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ...exceptions import PincerError
from ...utils import APIObject


@dataclass
class DiscordError(PincerError, APIObject):
    """Represents an error event in the Discord Gateway.

    Attributes
    ----------
    code: :class:`int`
        The RPC error code.
    message: :class:`str`
        The error description.
    """
    code: int
    message: str



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from .emoji import Emoji


@dataclass
class Reaction(APIObject):
    """Represents a Discord Reaction object

    Attributes
    ----------
    count: :class:`int`
        Times this emoji has been used to react
    me: :class:`bool`
        Whether the current user reacted using this emoji
    emoji: :class:`~pincer.objects.message.emoji.Emoji`
        Emoji information
    """
    count: int
    me: bool
    emoji: Emoji



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild is deleted"""

from ..core.dispatch import GatewayDispatch
from ..objects.guild import UnavailableGuild
from ..utils.conversion import construct_client_dict


async def guild_delete_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_delete`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild delete event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.guild.UnavailableGuild`]]
        ``on_guild_delete`` and an ``UnavailableGuild``
    """
    guild = UnavailableGuild.from_dict(construct_client_dict(self, payload.data))

    if guild.id in self.guilds.key():
        self.guilds.pop(guild.id)

    return "on_guild_delete", [guild]


def export():
    return guild_delete_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.


from dataclasses import dataclass

from ...objects.message import UserMessage
from ...utils.api_object import APIObject
from ...utils.snowflake import Snowflake


@dataclass
class NotificationCreateEvent(APIObject):
    """
    Represents a notification

    Attributes
    ----------
    channel_id : :class:`Snowflake`
        id of channel where notification occurred

    message : :class:`UserMessage`
        message that generated this notification

    icon_url : :class:`str`
        icon url of the notification

    title : :class:`str`
        title of the notification

    body : :class:`str`
        body of the notification
    """

    channel_id: Snowflake
    message: UserMessage
    icon_url: str
    title: str
    body: str



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..core.dispatch import GatewayDispatch


async def payload_middleware(
    payload: GatewayDispatch
) -> Tuple[str, List[GatewayDispatch]]:
    """Invoked when basically anything is received from gateway.

    Parameters
    ----------
    payload : :class:`pincer.core.dispatch.GatewayDispatch`
        The data received from the ready event.


    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.core.dispatch.GatewayDispatch`]]
        ``on_payload`` and a ``payload``
    """
    return "on_payload", [payload]


def export():
    return payload_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING, APINullable

if TYPE_CHECKING:
    from ...utils.snowflake import Snowflake


@dataclass
class IntegrationDeleteEvent(APIObject):
    """Sent when an integration is deleted.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        integration id
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        id of the guild
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        id of the bot/OAuth2 application for this discord integration
    """
    id: Snowflake
    guild_id: Snowflake
    application_id: APINullable[Snowflake] = MISSING


@dataclass
class IntegrationCreateEvent(APIObject):
    """
    Sent when an integration is created.

    Attributes
    ----------
    id : :class:`Snowflake`
        integration id

    guild_id : :class:`Snowflake`
        id of the guild

    application_id : APINullable[:class:`Snowflake`]
        id of the bot/OAuth2 application for this discord integration
    """
    id: Snowflake
    guild_id: Snowflake
    application_id: APINullable[Snowflake] = MISSING


@dataclass
class IntegrationUpdateEvent(APIObject):
    """
    Sent when an integration is updated.

    Attributes
    ----------
    id : :class:`Snowflake`
        integration id

    guild_id : :class:`Snowflake`
        id of the guild

    application_id : APINullable[:class:`Snowflake`]
        id of the bot/OAuth2 application for this discord integration
    """
    id: Snowflake
    guild_id: Snowflake
    application_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .application import Application
from .command import (
    AppCommandInteractionDataOption,
    AppCommandOptionChoice, AppCommandOption, AppCommand,
    ClientCommandStructure
)
from .command_types import AppCommandType, AppCommandOptionType
from .intents import Intents
from .interaction_base import CallbackType, InteractionType, MessageInteraction
from .interaction_flags import InteractionFlags
from .interactions import ResolvedData, InteractionData, Interaction
from .select_menu import SelectOption, SelectMenu
from .session_start_limit import SessionStartLimit
from .throttle_scope import ThrottleScope
from .throttling import ThrottleInterface, DefaultThrottleHandler


__all__ = (
    "AppCommand", "AppCommandInteractionDataOption",
    "AppCommandOption", "AppCommandOptionChoice", "AppCommandOptionType",
    "AppCommandType", "Application", "CallbackType", "ClientCommandStructure",
    "DefaultThrottleHandler", "Intents", "Interaction", "InteractionData",
    "InteractionFlags", "InteractionType", "MessageInteraction",
    "ResolvedData", "SelectMenu", "SelectOption", "SessionStartLimit",
    "ThrottleInterface", "ThrottleScope"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import os
from dataclasses import dataclass
from io import BytesIO
from typing import Any
from typing import TYPE_CHECKING

from ...utils import APIObject

if TYPE_CHECKING:
    from typing import Optional


PILLOW_IMPORT = True

IMAGE_TYPE = Any

try:
    from PIL.Image import Image

    if TYPE_CHECKING:
        IMAGE_TYPE = Image
except (ModuleNotFoundError, ImportError):
    PILLOW_IMPORT = False


@dataclass
class File(APIObject):
    """A file that is prepared by the user to be send to the discord
    API.

    Attributes
    ----------
    content: :class:`bytes`
        File bytes.
    filename: :class:`str`
        The name of the file when its uploaded to discord.
    """

    content: bytes
    filename: str

    @classmethod
    def from_file(cls, filepath: str, filename: str = None) -> File:
        """Make a ``File`` object from a file stored locally.

        Parameters
        ----------
        filepath: :class:`str`
            The path to the file you want to send. Must be string. The file's
            name in the file path is used as the name when uploaded to discord
            by default.

        filename: :class:`str`
            The name of the file. Will override the default name.
            |default| ``os.path.basename(filepath)``
        """
        with open(filepath, "rb") as data:
            file = data.read()

        return cls(
            content=file,
            filename=filename or os.path.basename(filepath)
        )

    @classmethod
    def from_pillow_image(
            cls,
            img: IMAGE_TYPE,
            filename: str,
            image_format: Optional[str] = None,
            **kwargs
    ) -> File:
        """Creates a file object from a PIL image
        Supports GIF, PNG, JPEG, and WEBP.

        Parameters
        ----------
        img: :class:`~pil:PIL.Image.Image`
            Pillow image object.
        filename:
            The filename to be used when uploaded to discord. The extension is
            used as image_format unless otherwise specified.
        image_format:
            The image_format to be used if you want to override the file
            extension.

        Returns
        -------
        :class:`~pincer.objects.message.file.File`
            The new file object.

        Raises
        ------
        ModuleNotFoundError:
            ``Pillow`` is not installed
        """
        if not PILLOW_IMPORT:
            raise ModuleNotFoundError(
                "The `Pillow` library is required for sending and converting "
                "pillow images,"
            )

        if image_format is None:
            image_format = os.path.splitext(filename)[1][1:]

            if image_format == "jpg":
                image_format = "jpeg"

        # https://stackoverflow.com/questions/33101935/convert-pil-image-to-byte-array
        # Credit goes to second answer
        img_byte_arr = BytesIO()
        img.save(img_byte_arr, format=image_format)
        img_bytes = img_byte_arr.getvalue()

        return cls(
            content=img_bytes,
            filename=filename
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import APINullable, MISSING

if TYPE_CHECKING:
    from typing import List

    from ..guild.channel import Channel
    from ..guild.thread import ThreadMember
    from ...utils.snowflake import Snowflake


@dataclass
class ThreadListSyncEvent(APIObject):
    """Sent when the current user gains access to a channel.

    Attributes
    ----------
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the guild
    threads: List[:class:`~pincer.objects.guild.channel.Channel`]
        All active threads in the given channels that
        the current user can access
    members: List[:class:`~pincer.objects.guild.thread.ThreadMember`]
        All thread member objects from the synced threads for
        the current user, indicating which threads the current
        user has been added to
    channel_ids: APINullable[List[:class:`~pincer.utils.snowflake.Snowflake`]]
        The parent channel ids whose threads are being synced.
        If omitted, then threads were synced for the entire guild.
        This array may contain channel_ids that have no active
        threads as well, so you know to clear that data.
    """
    guild_id: Snowflake
    threads: List[Channel]
    members: List[ThreadMember]

    channel_ids: APINullable[List[Snowflake]] = MISSING


@dataclass
class ThreadMembersUpdateEvent(APIObject):
    """Sent when anyone is added to or removed from a thread.
    If the current user does not have the `GUILD_MEMBERS`
    Gateway Intent, then this event will only be sent if
    the current user was added to or removed from the thread.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the thread
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the guild
    member_count: :class:`int`
        The approximate number of members in the thread, capped at 50
    added_members: APINullable[List[:class:`~pincer.objects.guild.thread.ThreadMember`]]
        The users who were added to the thread
    removed_member_ids: APINullable[List[:class:`~pincer.utils.snowflake.Snowflake`]]
        The id of the users who were removed from the thread
    """
    # noqa: E501
    id: Snowflake
    guild_id: Snowflake
    member_count: int

    added_members: APINullable[List[ThreadMember]] = MISSING
    removed_member_ids: APINullable[List[Snowflake]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Optional

    from ..user.user import User
    from ..guild.guild import Guild
    from ..guild.channel import Channel
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


class WebhookType(IntEnum):
    """Represents the type of a webhook.

    Attributes
    ----------
    INCOMING:
        Incoming Webhooks can post messages to channels with a
        generated token.
    CHANNEL_FOLLOWER:
        Channel Follower Webhooks are internal webhooks used with
        Channel Following to post new messages into channels.
    APPLICATION:
        Application webhooks are webhooks used with Interactions
    """
    INCOMING = 1
    CHANNEL_FOLLOWER = 2
    APPLICATION = 3


@dataclass
class Webhook(APIObject):
    """Represents a Discord channel webhook.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the webhook
    type: :class:`~pincer.objects.guild.webhook.WebhookType`
        The type of the webhook
    channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The channel id this webhook is for, if any
    name: Optional[:class:`str`]
        The default name of the webhook
    avatar: Optional[:class:`str`]
        The default user avatar hash of the webhook
    application_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The bot/OAuth2 application that created this webhook
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        The user this webhook was created by
        (not returned when getting a webhook with its token)
    token: APINullable[:class:`str`]
        The secure token of the webhook
        (returned for Incoming Webhooks)
    source_guild: APINullable[:class:`~pincer.objects.guild.guild.Guild`]
        The guild of the channel that this webhook is following
        (returned for Channel Follower Webhooks)
    source_channel: APINullable[:class:`~pincer.objects.guild.channel.Channel`]
        The channel that this webhook is following
        (returned for Channel Follower Webhooks)
    url: APINullable[:class:`str`]
        The url used for executing the webhook
        (returned by the webhooks OAuth2 flow)
    guild_id: APINullable[Optional[:class:`~pincer.objects.guild.guild.Guild`]]
        The guild id this webhook is for, if any
    """
    id: Snowflake
    type: WebhookType

    channel_id: Optional[Snowflake] = None
    name: Optional[str] = None
    avatar: Optional[str] = None
    application_id: Optional[Snowflake] = None

    user: APINullable[User] = MISSING
    token: APINullable[str] = MISSING
    source_guild: APINullable[Guild] = MISSING
    source_channel: APINullable[Channel] = MISSING
    url: APINullable[str] = MISSING

    guild_id: APINullable[Optional[Snowflake]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when the client receives a notification
(mention or new message in eligible channels)
"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.notification import NotificationCreateEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def notification_create_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_notification_create`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the notification create event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.notification.NotificationCreateEvent`]]
        ``on_notification_create`` and a ``NotificationCreateEvent``
    """
    channel_id: int = payload.data.get("channel_id")
    payload.data["message"]["channel_id"] = channel_id
    return "on_notification_create", [
        NotificationCreateEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return notification_create_middleware



"""
Pincer library
====================
An asynchronous python API wrapper meant to replace discord.py

Copyright Pincer 2021
Full MIT License can be found in `LICENSE` at the project root.
"""

from typing import NamedTuple, Literal, Optional

from ._config import GatewayConfig
from .client import event_middleware, Client, Bot
from .commands import command, ChatCommandHandler
from .exceptions import (
    PincerError, UnhandledException, NoExportMethod,
    CogError, CogNotFound, CogAlreadyExists, NoValidSetupMethod,
    TooManySetupArguments, NoCogManagerReturnFound, CommandError,
    CommandCooldownError, CommandIsNotCoroutine, CommandAlreadyRegistered,
    CommandDescriptionTooLong, TooManyArguments, InvalidArgumentAnnotation,
    CommandReturnIsEmpty, InvalidCommandGuild, InvalidCommandName,
    InvalidEventName, InvalidUrlError, EmbedFieldError, TaskError,
    TaskAlreadyRunning, TaskCancelError, TaskIsNotCoroutine, TaskInvalidDelay,
    DispatchError, DisallowedIntentsError, InvalidTokenError, HeartbeatError,
    UnavailableGuildError, HTTPError, NotModifiedError, BadRequestError,
    UnauthorizedError, ForbiddenError, NotFoundError, MethodNotAllowedError,
    RateLimitError, GatewayError, ServerError
)
from .objects import Intents
from .utils import Choices, Descripted

__package__ = "pincer"
__title__ = "Pincer library"
__description__ = "Discord API wrapper rebuild from scratch."
__author__ = "Sigmanificient, Arthurdw"
__email__ = "contact@pincer.dev"
__license__ = "MIT"

ReleaseType = Optional[Literal["alpha", "beta", "candidate", "final", "dev"]]


class VersionInfo(NamedTuple):
    """A Class representing the version of the Pincer library."""

    major: int
    minor: int
    micro: int

    release_level: ReleaseType = None
    serial: int = 0

    def __repr__(self) -> str:
        return f"{self.major}.{self.minor}.{self.micro}" + (
            f"-{self.release_level}{self.serial}"
            * (self.release_level is not None)
        )


version_info = VersionInfo(0, 11, 5)
__version__ = repr(version_info)

__all__ = (
    "BadRequestError", "Bot", "ChatCommandHandler", "Choices",
    "Client", "CogAlreadyExists", "CogError", "CogNotFound",
    "CommandAlreadyRegistered", "CommandCooldownError",
    "CommandDescriptionTooLong", "CommandError", "CommandIsNotCoroutine",
    "CommandReturnIsEmpty", "Descripted", "DisallowedIntentsError",
    "DispatchError", "EmbedFieldError", "ForbiddenError", "GatewayConfig",
    "GatewayError", "HTTPError", "HeartbeatError", "Intents",
    "InvalidArgumentAnnotation", "InvalidCommandGuild", "InvalidCommandName",
    "InvalidEventName", "InvalidTokenError", "InvalidUrlError",
    "MethodNotAllowedError", "NoCogManagerReturnFound", "NoExportMethod",
    "NoValidSetupMethod", "NotFoundError", "NotModifiedError", "PincerError",
    "RateLimitError", "ServerError", "TaskAlreadyRunning", "TaskCancelError",
    "TaskError", "TaskInvalidDelay", "TaskIsNotCoroutine", "TooManyArguments",
    "TooManySetupArguments", "UnauthorizedError", "UnavailableGuildError",
    "UnhandledException", "__author__", "__email__", "__package__",
    "__title__", "__version__", "command", "event_middleware"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

from sys import modules
from typing import (
    TYPE_CHECKING, TypeVar, Callable, Coroutine, Any, Union, Literal, Optional
)

from pincer.exceptions import InvalidArgumentAnnotation

if TYPE_CHECKING:
    from typing import Tuple


class MissingType:
    """Type class for missing attributes and parameters."""

    def __repr__(self):
        return "<MISSING>"

    def __bool__(self) -> bool:
        return False


MISSING = MissingType()


T = TypeVar('T')


# Represents a value which is optionally returned from the API
APINullable = Union[T, MissingType]


# Represents a coroutine.
Coro = TypeVar("Coro", bound=Callable[..., Coroutine[Any, Any, Any]])


Choices = Literal


choice_value_types = (str, int, float)

CheckFunction = Optional[Callable[[Any], bool]]


class Singleton(type):
    # Thanks to this stackoverflow answer (method 3):
    # https://stackoverflow.com/q/6760685/12668716
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton,
                cls
            ).__call__(*args, **kwargs)
        return cls._instances[cls]


class TypeCache(metaclass=Singleton):
    cache = {}

    def __init__(self):
        # Register all known types to the cache. This gets used later
        # to auto-convert the properties to their desired type.
        lcp = modules.copy()
        for module in lcp:
            if not module.startswith("pincer"):
                continue

            TypeCache.cache.update(lcp[module].__dict__)


class _TypeInstanceMeta(type):
    def __getitem__(cls, args: Tuple[T, str]):
        if not isinstance(args, tuple) or len(args) != 2:
            raise InvalidArgumentAnnotation(
                "Descripted arguments must be a tuple of length 2. "
                "(if you are using this as the indented type, just "
                "pass two arguments)"
            )

        return cls(*args)


class Descripted(metaclass=_TypeInstanceMeta):
    # TODO: Write example & more docs
    """Description type."""

    def __init__(self, key: Any, description: str):
        if not isinstance(description, str):
            raise RuntimeError(
                "The description value must always be a string!"
            )

        self.key = key
        self.description = description



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
Sent when the user clicks a Rich Presence join invite in chat
to join a game.
"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.activity import ActivityJoinEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def activity_join_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_activity_join`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the activity join event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`ActivityJoinEvent`]]
        ``on_activity_join`` and an ``ActivityJoinEvent``
    """
    return "on_activity_join", [
        ActivityJoinEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return activity_join_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .attachment import Attachment
from .button import ButtonStyle, Button
from .component import MessageComponent
from .context import MessageContext
from .embed import (
    Embed, EmbedField, EmbedImage, EmbedAuthor, EmbedProvider, EmbedThumbnail,
    EmbedVideo, EmbedFooter
)
from .emoji import Emoji
from .file import File
from .message import Message
from .reaction import Reaction
from .reference import MessageReference
from .sticker import (
    StickerType, StickerFormatType, Sticker, StickerItem, StickerPack
)
from .user_message import (
    MessageActivityType, MessageFlags, MessageType, MessageActivity,
    AllowedMentionTypes, UserMessage, AllowedMentions
)

__all__ = (
    "AllowedMentionTypes", "AllowedMentions", "Attachment", "Button",
    "ButtonStyle", "Embed", "EmbedAuthor", "EmbedField", "EmbedFooter",
    "EmbedImage", "EmbedProvider", "EmbedThumbnail", "EmbedVideo", "Emoji",
    "File", "Message", "MessageActivity", "MessageActivityType",
    "MessageComponent", "MessageContext", "MessageFlags", "MessageReference",
    "MessageType", "Reaction", "Sticker", "StickerFormatType", "StickerItem",
    "StickerPack", "StickerType", "UserMessage"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when an invite is deleted"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.invite import InviteDeleteEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def invite_delete_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_invite_delete``,

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the invite delete event

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.events.invite.InviteDeleteEvent`]]
        ``on_invite_delete`` and an ``InviteDeleteEvent``
    """
    return "on_invite_delete", [
        InviteDeleteEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return invite_delete_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
from asyncio import iscoroutinefunction, run, ensure_future
from collections import defaultdict
from importlib import import_module
from inspect import isasyncgenfunction
from typing import Any, Dict, List, Optional, Tuple, Union, overload
from typing import TYPE_CHECKING

from . import __package__
from .commands import ChatCommandHandler
from .core import HTTPClient
from .core.gateway import Dispatcher
from .exceptions import (
    InvalidEventName, TooManySetupArguments, NoValidSetupMethod,
    NoCogManagerReturnFound, CogAlreadyExists, CogNotFound
)
from .middleware import middleware
from .objects import (
    Role, Channel, DefaultThrottleHandler, User, Guild, Intents
)
from .utils.event_mgr import EventMgr
from .utils.extraction import get_index
from .utils.insertion import should_pass_cls
from .utils.signature import get_params
from .utils.types import CheckFunction
from .utils.types import Coro

if TYPE_CHECKING:
    from .objects.app import AppCommand
    from .utils.snowflake import Snowflake
    from .core.dispatch import GatewayDispatch
    from .objects.app.throttling import ThrottleInterface

_log = logging.getLogger(__package__)

MiddlewareType = Optional[Union[Coro, Tuple[str, List[Any], Dict[str, Any]]]]

_event = Union[str, Coro]
_events: Dict[str, Optional[Union[List[_event], _event]]] = defaultdict(list)


def event_middleware(call: str, *, override: bool = False):
    """Middleware are methods which can be registered with this decorator.
    These methods are invoked before any ``on_`` event.
    As the ``on_`` event is the final call.

    A default call exists for all events, but some might already be in
    use by the library.

    If you know what you are doing, you can override these default
    middleware methods by passing the override parameter.

    The method to which this decorator is registered must be a coroutine,
    and it must return a tuple with the following format:

    .. code-block:: python

        tuple(
            key for next middleware or final event [str],
            args for next middleware/event which will be passed as *args
                [list(Any)],
            kwargs for next middleware/event which will be passed as
                **kwargs [dict(Any)]
        )

    Two parameters are passed to the middleware. The first parameter is
    the current socket connection with and the second one is the payload
    parameter which is of type :class:`~.core.dispatch.GatewayDispatch`.
    This contains the response from the discord API.

    :Implementation example:

    .. code-block:: pycon

        >>> @event_middleware("ready", override=True)
        >>> async def custom_ready(_, payload: GatewayDispatch):
        >>>     return "on_ready", [
        >>>         User.from_dict(payload.data.get("user"))
        >>>     ]

        >>> @Client.event
        >>> async def on_ready(bot: User):
        >>>     print(f"Signed in as {bot}")

    Parameters
    ----------
    call : :class:`str`
        The call that the function should tie to
    override : :class:`bool`
        If it should override default middleware,
        usually shouldn't be used |default| :data:`False`
    """

    def decorator(func: Coro):
        if override:
            _log.warning(
                "Middleware overriding has been enabled for `%s`."
                " This might cause unexpected behavior.", call
            )

        if not override and callable(_events.get(call)):
            raise RuntimeError(
                f"Middleware event with call `{call}` has "
                "already been registered"
            )

        async def wrapper(cls, payload: GatewayDispatch):
            _log.debug("`%s` middleware has been invoked", call)

            return await (
                func(cls, payload)
                if should_pass_cls(func)
                else func(payload)
            )

        _events[call] = wrapper
        return wrapper

    return decorator


for event, middleware_ in middleware.items():
    event_middleware(event)(middleware_)


class Client(Dispatcher):
    """The client is the main instance which is between the programmer
    and the discord API.

    This client represents your bot.

    Attributes
    ----------
    bot: :class:`~objects.user.user.User`
        The user object of the client
    received_message: :class:`str`
        The default message which will be sent when no response is given.
    http: :class:`~core.http.HTTPClient`
        The http client used to communicate with the discord API

    Parameters
    ----------
    token : :class:`str`
        The token to login with your bot from the developer portal
    received : Optional[:class:`str`]
        The default message which will be sent when no response is given.
        |default| :data:`None`
    intents : Optional[:class:`~objects.app.intents.Intents`]
        The discord intents to use |default| :data:`None`
    throttler : Optional[:class:`~objects.app.throttling.ThrottleInterface`]
        The cooldown handler for your client,
        defaults to :class:`~.objects.app.throttling.DefaultThrottleHandler`
        *(which uses the WindowSliding technique)*.
        Custom throttlers must derive from
        :class:`~pincer.objects.app.throttling.ThrottleInterface`.
        |default| :class:`~pincer.objects.app.throttling.DefaultThrottleHandler`
    """

    def __init__(
            self,
            token: str, *,
            received: str = None,
            intents: Intents = None,
            throttler: ThrottleInterface = DefaultThrottleHandler,
            reconnect: bool = True,
    ):
        super().__init__(
            token,
            handlers={
                # Gets triggered on all events
                -1: self.payload_event_handler,
                # Use this event handler for opcode 0.
                0: self.event_handler
            },
            intents=intents or Intents.NONE,
            reconnect=reconnect,
        )

        self.bot: Optional[User] = None
        self.received_message = received or "Command arrived successfully!"
        self.http = HTTPClient(token)
        self.throttler = throttler
        self.event_mgr = EventMgr()
        # TODO: Document guild prop
        # The guild value is only registered if the GUILD_MEMBERS
        # intent is enabled.
        self.guilds: Dict[Snowflake, Optional[Guild]] = {}
        ChatCommandHandler.managers[self.__module__] = self

    @property
    def chat_commands(self) -> List[str]:
        """List[:class:`str`]: List of chat commands

        Get a list of chat command calls which have been registered in
        the :class:`~pincer.commands.ChatCommandHandler`\\.
        """
        return list(map(
            lambda cmd: cmd.app.name,
            ChatCommandHandler.register.values()
        ))

    @staticmethod
    def event(coroutine: Coro):
        """A decorator to register a Discord gateway event listener.
        This event will get called when the client receives a new event
        update from Discord which matches the event name.

        The event name gets pulled from your method name, and this must
        start with ``on_``.
        This forces you to write clean and consistent code.

        This decorator can be used in and out of a class, and all
        event methods must be coroutines. *(async)*

        :Example usage:

        .. code-block:: pycon

            >>> # Function based
            >>> from pincer import Client
            >>>
            >>> client = Client("token")
            >>>
            >>> @client.event
            >>> async def on_ready():
            ...     print(f"Signed in as {client.bot}")
            >>>
            >>> if __name__ == "__main__":
            ...     client.run()

        .. code-block :: pycon

            >>> # Class based
            >>> from pincer import Client
            >>>
            >>> class MyClient(Client):
            ...     @Client.event
            ...     async def on_ready(self):
            ...         print(f"Signed in as {self.bot}")
            >>>
            >>> if __name__ == "__main__":
            ...     MyClient("token").run()

        Raises
        ------
        TypeError
            If the function is not a coro
        InvalidEventName
            If the function name is not a valid event (on_x)
        """
        if not iscoroutinefunction(coroutine) \
                and not isasyncgenfunction(coroutine):
            raise TypeError(
                "Any event which is registered must be a coroutine function"
            )

        name: str = coroutine.__name__.lower()

        if not name.startswith("on_"):
            raise InvalidEventName(
                f"The event named `{name}` must start with `on_`"
            )

        if name == "on_command_error" and _events.get(name):
            raise InvalidEventName(
                f"The `{name}` event can only exist once. This is because "
                "it gets treated as a command and can have a response."
            )

        _events[name].append(coroutine)
        return coroutine

    @staticmethod
    def get_event_coro(name: str) -> List[Optional[Coro]]:
        """get the coroutine for an event

        Parameters
        ----------
        name : :class:`str`
            name of the event
        """
        calls = _events.get(name.strip().lower())

        return [] if not calls else list(filter(
            lambda call: iscoroutinefunction(call) or isasyncgenfunction(call),
            calls
        ))

    def load_cog(self, path: str, package: Optional[str] = None):
        """Load a cog from a string path, setup method in COG may
        optionally have a first argument which will contain the client!

        :Example usage:

        run.py

        .. code-block:: pycon

            >>> from pincer import Client
            >>>
            >>> class MyClient(Client):
            ...     def __init__(self, *args, **kwargs):
            ...         self.load_cog("cogs.say")
            ...         super().__init__(*args, **kwargs)

        cogs/say.py

        .. code-block:: pycon

            >>> from pincer import command
            >>>
            >>> class SayCommand:
            ...     @command()
            ...     async def say(self, message: str) -> str:
            ...         return message
            >>>
            >>> setup = SayCommand

        Parameters
        ----------
        path : :class:`str`
            The import path for the cog.
        package : :class:`str`
            The package name for relative based imports.
            |default| :data:`None`
        """

        if ChatCommandHandler.managers.get(path):
            raise CogAlreadyExists(
                f"Cog `{path}` is trying to be loaded but already exists."
            )

        try:
            module = import_module(path, package=package)
        except ModuleNotFoundError:
            raise CogNotFound(f"Cog `{path}` could not be found!")

        setup = getattr(module, "setup", None)

        if not callable(setup):
            raise NoValidSetupMethod(
                f"`setup` method was expected in `{path}` but none was found!"
            )

        args, params = [], get_params(setup)

        if len(params) == 1:
            args.append(self)
        elif (length := len(params)) > 1:
            raise TooManySetupArguments(
                f"Setup method in `{path}` requested {length} arguments "
                f"but the maximum is 1!"
            )

        cog_manager = setup(*args)

        if not cog_manager:
            raise NoCogManagerReturnFound(
                f"Setup method in `{path}` didn't return a cog manager! "
                "(Did you forget to return the cog?)"
            )

        ChatCommandHandler.managers[path] = cog_manager

    @staticmethod
    def get_cogs() -> Dict[str, Any]:
        """Get a dictionary of all loaded cogs.

        The key/value pair is import path/cog class.

        Returns
        -------
        Dict[:class:`str`, Any]
            The dictionary of cogs
        """
        return ChatCommandHandler.managers

    async def unload_cog(self, path: str):
        """|coro|

        Unload an already loaded cog! This removes all of its commands!

        Parameters
        ----------
        path : :class:`str`
            The path to the cog

        Raises
        ------
        CogNotFound
            When the cog is not in that path
        """
        if not ChatCommandHandler.managers.get(path):
            raise CogNotFound(f"Cog `{path}` could not be found!")

        to_remove: List[AppCommand] = []

        for command in ChatCommandHandler.register.values():
            if not command:
                continue

            if command.call.__module__ == path:
                to_remove.append(command.app)

        await ChatCommandHandler(self).remove_commands(to_remove)

    @staticmethod
    def execute_event(calls: List[Coro], *args, **kwargs):
        """Invokes an event.

        Parameters
        ----------
        calls: :data:`~pincer.utils.types.Coro`
            The call (method) to which the event is registered.

        \\*args:
            The arguments for the event.

        \\*\\*kwargs:
            The named arguments for the event.
        """

        for call in calls:
            call_args = args
            if should_pass_cls(call):
                call_args = (
                    ChatCommandHandler.managers[call.__module__],
                    *args
                )

            ensure_future(call(*call_args, **kwargs))

    def run(self):
        """start the event listener"""
        self.start_loop()

    def __del__(self):
        if hasattr(self, 'http'):
            run(self.http.close())

    async def handle_middleware(
            self,
            payload: GatewayDispatch,
            key: str,
            *args,
            **kwargs
    ) -> Tuple[Optional[Coro], List[Any], Dict[str, Any]]:
        """|coro|

        Handles all middleware recursively. Stops when it has found an
        event name which starts with ``on_``.

        Returns a tuple where the first element is the final executor
        (so the event) its index in ``_events``.

        The second and third element are the ``*args``
        and ``**kwargs`` for the event.

        Parameters
        ----------
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The original payload for the event
        key : :class:`str`
            The index of the middleware in ``_events``

        Raises
        ------
        RuntimeError
            The return type must be a tuple
        RuntimeError
            Middleware has not been registered
        """
        ware: MiddlewareType = _events.get(key)
        next_call, arguments, params = ware, [], {}

        if iscoroutinefunction(ware):
            extractable = await ware(self, payload, *args, **kwargs)

            if not isinstance(extractable, tuple):
                raise RuntimeError(
                    f"Return type from `{key}` middleware must be tuple. "
                )

            next_call = get_index(extractable, 0, "")
            arguments = get_index(extractable, 1, [])
            params = get_index(extractable, 2, {})

        if next_call is None:
            raise RuntimeError(f"Middleware `{key}` has not been registered.")

        return (
            (next_call, arguments, params)
            if next_call.startswith("on_")
            else await self.handle_middleware(
                payload, next_call, *arguments, **params
            )
        )

    async def execute_error(
            self,
            error: Exception,
            name: str = "on_error",
            *args,
            **kwargs
    ):
        """|coro|

        Raises an error if no appropriate error event has been found.

        Parameters
        ----------
        error : :class:`Exception`
            The error that should be passed to the event
        name : :class:`str`
            the name of the event |default| ``on_error``

        Raises
        ------
        error
            if ``call := self.get_event_coro(name)`` is :data:`False`
        """
        if calls := self.get_event_coro(name):
            self.execute_event(calls, error, *args, **kwargs)
        else:
            raise error

    async def process_event(self, name: str, payload: GatewayDispatch):
        """|coro|

        Processes and invokes an event and its middleware

        Parameters
        ----------
        name : :class:`str`
            The name of the event, this is also the filename in the
            middleware directory.
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The payload sent from the Discord gateway, this contains the
            required data for the client to know what event it is and
            what specifically happened.
        """
        try:
            key, args, kwargs = await self.handle_middleware(payload, name)

            self.event_mgr.process_events(key, *args)

            if calls := self.get_event_coro(key):
                self.execute_event(calls, *args, **kwargs)

        except Exception as e:
            await self.execute_error(e)

    async def event_handler(self, _, payload: GatewayDispatch):
        """|coro|

        Handles all payload events with opcode 0.

        Parameters
        ----------
        _ :
            Socket param, but this isn't required for this handler. So
            its just a filler parameter, doesn't matter what is passed.
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The payload sent from the Discord gateway, this contains the
            required data for the client to know what event it is and
            what specifically happened.
        """
        await self.process_event(payload.event_name.lower(), payload)

    async def payload_event_handler(self, _, payload: GatewayDispatch):
        """|coro|

        Special event which activates on_payload event!

        Parameters
        ----------
        _ :
            Socket param, but this isn't required for this handler. So
            its just a filler parameter, doesn't matter what is passed.
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The payload sent from the Discord gateway, this contains the
            required data for the client to know what event it is and
            what specifically happened.
        """
        await self.process_event("payload", payload)
    
    @overload
    async def create_guild(
        self,
        *,
        name: str,
        region: Optional[str] = None,
        icon: Optional[str] = None,
        verification_level: Optional[int] = None,
        default_message_notifications: Optional[int] = None,
        explicit_content_filter: Optional[int] = None,
        roles: Optional[List[Role]] = None,
        channels: Optional[List[Channel]] = None,
        afk_channel_id: Optional[Snowflake] = None,
        afk_timeout: Optional[int] = None,
        system_channel_id: Optional[Snowflake] = None,
        system_channel_flags: Optional[int] = None
    ) -> Guild:
        """Creates a guild.

        Parameters
        ----------
        name : :class:`str`
            Name of the guild (2-100 characters)
        region : Optional[:class:`str`]
            Voice region id (deprecated) |default| :data:`None`
        icon : Optional[:class:`str`]
            base64 128x128 image for the guild icon |default| :data:`None`
        verification_level : Optional[:class:`int`]
            Verification level |default| :data:`None`
        default_message_notifications : Optional[:class:`int`]
            Default message notification level |default| :data:`None`
        explicit_content_filter : Optional[:class:`int`]
            Explicit content filter level |default| :data:`None`
        roles : Optional[List[:class:`~pincer.objects.guild.role.Role`]]
            New guild roles |default| :data:`None`
        channels : Optional[List[:class:`~pincer.objects.guild.channel.Channel`]]
            New guild's channels |default| :data:`None`
        afk_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID for AFK channel |default| :data:`None`
        afk_timeout : Optional[:class:`int`]
            AFK timeout in seconds |default| :data:`None`
        system_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where guild notices such as welcome
            messages and boost events are posted |default| :data:`None`
        system_channel_flags : Optional[:class:`int`]
            System channel flags |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
            The created guild
        """
        ...
    
    async def create_guild(self, name: str, **kwargs) -> Guild:
        g = await self.http.post("guilds", data={"name": name, **kwargs})
        return await self.get_guild(g['id'])

    async def wait_for(
            self,
            event_name: str,
            check: CheckFunction = None,
            timeout: Optional[float] = None
    ):
        """
        Parameters
        ----------
        event_name : str
            The type of event. It should start with `on_`. This is the same
            name that is used for @Client.event.
        check : CheckFunction
            This function only returns a value if this return true.
        timeout: Optional[float]
            Amount of seconds before timeout.

        Returns
        ------
        Any
            What the Discord API returns for this event.
        """
        return await self.event_mgr.wait_for(event_name, check, timeout)

    def loop_for(
            self,
            event_name: str,
            check: CheckFunction = None,
            iteration_timeout: Optional[float] = None,
            loop_timeout: Optional[float] = None
    ):
        """
        Parameters
        ----------
        event_name : str
            The type of event. It should start with `on_`. This is the same
            name that is used for @Client.event.
        check : Callable[[Any], bool], default=None
            This function only returns a value if this return true.
        iteration_timeout: Union[float, None]
            Amount of seconds before timeout. Timeouts are for each loop.
        loop_timeout: Union[float, None]
            Amount of seconds before the entire loop times out. The generator
            will only raise a timeout error while it is waiting for an event.

        Yields
        ------
        Any
            What the Discord API returns for this event.
        """
        return self.event_mgr.loop_for(
            event_name,
            check,
            iteration_timeout,
            loop_timeout
        )

    async def get_guild(self, guild_id: int) -> Guild:
        """|coro|

        Fetch a guild object by the guild identifier.

        Parameters
        ----------
        guild_id : :class:`int`
            The id of the guild which should be fetched from the Discord
            gateway.
        """
        return await Guild.from_id(self, guild_id)

    async def get_user(self, _id: int) -> User:
        """|coro|

        Fetch a User from its identifier

        Parameters
        ----------
        _id : :class:`int`
            The id of the user which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.user.user.User`
            The user object.
        """
        return await User.from_id(self, _id)

    async def get_role(self, guild_id: int, role_id: int) -> Role:
        """|coro|

        Fetch a role object by the role identifier.

        guild_id: :class:`int`
            The guild in which the role resides.

        role_id: :class:`int`
            The id of the guild which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.guild.role.Role`
            A Role object.
        """
        return await Role.from_id(self, guild_id, role_id)

    async def get_channel(self, _id: int) -> Channel:
        """Fetch a Channel from its identifier.

        _id: :class:`int`
            The id of the user which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            A Channel object.
        """
        return await Channel.from_id(self, _id)


Bot = Client



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from typing import overload, TYPE_CHECKING

from .channel import Channel
from .member import GuildMember
from ...exceptions import UnavailableGuildError
from ...utils.api_object import APIObject
from ...utils.conversion import construct_client_dict
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Union

    from .features import GuildFeature
    from .role import Role
    from .stage import StageInstance
    from .welcome_screen import WelcomeScreen
    from ..events.presence import PresenceUpdateEvent
    from ..message.emoji import Emoji
    from ..message.sticker import Sticker
    from ..user.voice_state import VoiceState
    from ...client import Client
    from ...utils.timestamp import Timestamp
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


class PremiumTier(IntEnum):
    """Represents the boost tier of a guild.
    Attributes
    ----------
    NONE:
        Guild has not unlocked any Server Boost perks.
    TIER_1:
        Guild has unlocked Server Boost level 1 perks.
    TIER_2:
        Guild has unlocked Server Boost level 2 perks.
    TIER_3:
        Guild has unlocked Server Boost level 3 perks.
    """
    NONE = 0
    TIER_1 = 1
    TIER_2 = 2
    TIER_3 = 3


class GuildNSFWLevel(IntEnum):
    """Represents the NSFW level of a guild.
    Attributes
    ----------
    DEFAULT:
        Default NSFW level.
    EXPLICIT:
        Explicit NSFW level.
    SAFE:
        SAFE NSFW level.
    AGE_RESTRICTED:
        Age restricted NSFW level.
    """
    DEFAULT = 0
    EXPLICIT = 1
    SAFE = 2
    AGE_RESTRICTED = 3


class ExplicitContentFilterLevel(IntEnum):
    """Represents the filter content level of a guild.
    Attributes
    ----------
    DISABLED:
        Media content will not be scanned.
    MEMBERS_WITHOUT_ROLES:
        Media content sent by members without roles will be scanned.
    ALL_MEMBERS:
        Media content sent by all members will be scanned.
    """
    DISABLED = 0
    MEMBERS_WITHOUT_ROLES = 1
    ALL_MEMBERS = 2


class MFALevel(IntEnum):
    """Represents the multi factor authentication level of a guild.
    Attributes
    ----------
    NONE:
        Guild has no MFA/2FA requirement for moderation actions.
    ELEVATED:
        Guild has a 2FA requirement for moderation actions
    """
    NONE = 0
    ELEVATED = 1


class VerificationLevel(IntEnum):
    """Represents the verification level of a guild.
    Attributes
    ----------
    NONE:
        Unrestricted.
    LOW:
        Must have verified email on account.
    MEDIUM:
        Must be registered on Discord for longer than 5 minutes.
    HIGH:
        Must be a member of the server for longer than 10 minutes.
    VERY_HIGH:
        Must have a verified phone number.
    """
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    VERY_HIGH = 4


class DefaultMessageNotificationLevel(IntEnum):
    """Represents the default message notification level of a guild.
    Attributes
    ----------
    ALL_MESSAGES:
        Members will receive notifications for all messages by default.
    ONLY_MENTIONS:
        Members will receive notifications only for messages that @mention them by default.
    """
    # noqa: E501
    ALL_MESSAGES = 0
    ONLY_MENTIONS = 1


class SystemChannelFlags(IntEnum):
    """Represents the system channel flags of a guild.
    Attributes
    ----------
    SUPPRESS_JOIN_NOTIFICATIONS:
        Suppress member join notifications.
    SUPPRESS_PREMIUM_SUBSCRIPTIONS:
        Suppress server boost notifications.
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS:
        Suppress server setup tips.
    SUPPRESS_JOIN_NOTIFICATION_REPLIES:
        Hide member join sticker reply buttons
    """
    SUPPRESS_JOIN_NOTIFICATIONS = 1 << 0
    SUPPRESS_PREMIUM_SUBSCRIPTIONS = 1 << 1
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS = 1 << 2
    SUPPRESS_JOIN_NOTIFICATION_REPLIES = 1 << 3


@dataclass
class GuildPreview(APIObject):
    """Represents a guild preview.
    Attributes
    ----------
    id: :class:`Snowflake`
        The guild ID.
    name: :class:`str`
        The guild name.
    icon: :class:`str`
        The guild icon hash.
    splash: :class:`str`
        The guild splash hash.
    discovery_splash: :class:`str`
        The guild discovery splash hash.
    emojis: :class:`List[Emoji]`
        The guild emojis.
    features: :class:`List[GuildFeature]`
        The guild features.
    approximate_member_count: :class:`int`
        The approximate member count.
    approximate_presence_count: :class:`int`
        The approximate number of online members in this guild
    description: :class:`str`
        The guild description.
    """
    id: Snowflake
    name: str
    emojis: List[Emoji]
    features: List[GuildFeature]
    approximate_member_count: int
    approximate_presence_count: int

    icon: APINullable[str] = MISSING
    splash: APINullable[str] = MISSING
    discovery_splash: APINullable[str] = MISSING
    description: APINullable[str] = MISSING


@dataclass
class Guild(APIObject):
    """Represents a Discord guild/server in which your client resides.
    Attributes
    ----------
    afk_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of afk channel
    afk_timeout: :class:`int`
        Afk timeout in seconds
    application_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        Application id of the guild creator if it is bot-created
    banner: Optional[:class:`str`]
        Banner hash
    default_message_notifications: :class:`~pincer.objects.guild.guild.DefaultMessageNotificationLevel`
        Default message notifications level
    description: Optional[:class:`str`]
        The description of a Community guild
    discovery_splash: Optional[:class:`str`]
        Discovery splash hash;
        only present for guilds with the "DISCOVERABLE" feature
    emojis: List[:class:`~pincer.objects.message.emoji.Emoji`]
        Custom guild emojis
    explicit_content_filter: :class:`~pincer.objects.guild.guild.ExplicitContentFilterLevel`
        Explicit content filter level
    features: List[:class:`~pincer.objects.guild.features.GuildFeature`]
        Enabled guild features
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Guild id
    icon: Optional[:class:`str`]
        Icon hash
    mfa_level: :class:`~pincer.objects.guild.guild.MFALevel`
        Required MFA level for the guild
    name: :class:`str`
        Guild name (2-100 characters, excluding trailing and leading
        whitespace)
    nsfw_level: :class:`~pincer.objects.guild.guild.NSFWLevel`
        Guild NSFW level
    owner_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of owner
    preferred_locale: :class:`str`
        The preferred locale of a Community guild;
        used in server discovery and notices from Discord;
        defaults to "en-US"
    premium_tier: :class:`~pincer.objects.guild.guild.PremiumTier`
        Premium tier (Server Boost level)
    public_updates_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the channel where admins
        and moderators of Community guilds receive notices from Discord
    roles: List[:class:`~pincer.objects.guild.role.Role`]
        Roles in the guild
    rules_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the channel where Community guilds can display rules
        and/or guidelines
    splash: Optional[:class:`str`]
        Splash hash
    system_channel_flags: :class:`~pincer.objects.guild.guild.SystemChannelFlags`
        System channel flags
    system_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the channel where guild notices
        such as welcome messages and boost events are posted
    vanity_url_code: Optional[:class:`str`]
        The vanity url code for the guild
    verification_level: :class:`~pincer.objects.guild.guild.VerificationLevel`
        Verification level required for the guild
    approximate_member_count: APINullable[:class:`int`]
        Approximate number of members in this guild, returned from the
        `GET /guilds/<id>` endpoint when with_counts is true
    approximate_presence_count: APINullable[:class:`int`]
        Approximate number of non-offline members in this guild,
        returned from the `GET /guilds/<id>`
        endpoint when with_counts is true
    channels: APINullable[List[:class:`~pincer.objects.guild.channel.Channel`]]
        Channels in the guild
    icon_hash: APINullable[Optional[:class:`str`]]
        Icon hash, returned when in the template object
    joined_at: APINullable[:class:`~pincer.utils.timestamp.Timestamp`]
        When this guild was joined at
    large: APINullable[:class:`bool`]
        True if this is considered a large guild
    max_members: APINullable[:class:`int`]
        The maximum number of members for the guild
    max_presences: APINullable[Optional[:class:`int`]]
        The maximum number of presences for the guild
        (null is always returned, apart from the largest of guilds)
    max_video_channel_users: APINullable[:class:`int`]
        The maximum amount of users in a video channel
    members: APINullable[List[:class:`~pincer.objects.guild.member.GuildMember`]]
        Users in the guild
    member_count: APINullable[:class:`bool`]
        Total number of members in this guild
    nsfw: APINullable[:class:`bool`]
        Boolean if the server is NSFW
    owner: APINullable[:class:`bool`]
        True if the user is the owner of the guild
    permissions: APINullable[:class:`str`]
        Total permissions for the user in the guild
        (excludes overwrites)
    premium_subscription_count: APINullable[:class:`int`]
        The number of boosts this guild currently has
    presences: APINullable[List[:class:`~pincer.objects.events.presence.PresenceUpdateEvent`]]
        Presences of the members in the guild,
        will only include non-offline members if the size is greater
        than large threshold
    stage_instances: APINullable[List[:class:`~pincer.objects.guild.stage.StageInstance`]]
        Stage instances in the guild
    stickers: Optional[List[:class:`~pincer.objects.message.sticker.Sticker`]]
        Custom guild stickers
    region: APINullable[Optional[:class:`str`]]
        Voice region id for the guild (deprecated)
    threads: APINullable[List[:class:`~pincer.objects.guild.channel.Channel`]]
        All active threads in the guild that current user
        has permission to view
    unavailable: APINullable[:class:`bool`]
        True if this guild is unavailable due to an outage
    voice_states: APINullable[List[:class:`~pincer.objects.user.voice_state.VoiceState`]]
        States of members currently in voice channels;
        lacks the guild_id key
    widget_enabled: APINullable[:class:`bool`]
        True if the server widget is enabled
    widget_channel_id: APINullable[Optional[:class:`~pincer.utils.snowflake.Snowflake`]]
        The channel id that the widget will generate an invite to,
        or null if set to no invite
    welcome_screen: APINullable[:class:`~pincer.objects.guild.welcome_screen.WelcomeScreen`]
        The welcome screen of a Community guild, shown to new members,
        returned in an Invite's guild object
    """
    # noqa: E501
    afk_timeout: int
    default_message_notifications: DefaultMessageNotificationLevel
    emojis: List[Emoji]
    explicit_content_filter: ExplicitContentFilterLevel
    features: List[GuildFeature]
    id: Snowflake
    mfa_level: MFALevel
    name: str
    nsfw_level: GuildNSFWLevel
    owner_id: Snowflake
    preferred_locale: str
    premium_tier: PremiumTier
    roles: List[Role]
    system_channel_flags: SystemChannelFlags
    verification_level: VerificationLevel

    guild_scheduled_events: APINullable[List] = MISSING
    lazy: APINullable[bool] = MISSING
    premium_progress_bar_enabled: APINullable[bool] = MISSING
    guild_hashes: APINullable[Dict] = MISSING
    afk_channel_id: APINullable[Snowflake] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    embedded_activities: APINullable[List] = MISSING
    banner: APINullable[str] = MISSING
    description: APINullable[str] = MISSING
    discovery_splash: APINullable[str] = MISSING
    icon: APINullable[str] = MISSING
    public_updates_channel_id: APINullable[Snowflake] = MISSING
    rules_channel_id: APINullable[Snowflake] = MISSING
    splash: APINullable[str] = MISSING
    system_channel_id: APINullable[Snowflake] = MISSING
    vanity_url_code: APINullable[str] = MISSING

    application_command_counts: APINullable[Dict] = MISSING
    application_command_count: APINullable[int] = MISSING
    approximate_member_count: APINullable[int] = MISSING
    approximate_presence_count: APINullable[int] = MISSING
    channels: APINullable[List[Channel]] = field(default_factory=list)
    # TODO: Add type when type is known
    hub_type: APINullable[Any] = MISSING
    icon_hash: APINullable[Optional[str]] = MISSING
    joined_at: APINullable[Timestamp] = MISSING
    large: APINullable[bool] = MISSING
    max_members: APINullable[int] = MISSING
    max_presences: APINullable[Optional[int]] = MISSING
    max_video_channel_users: APINullable[int] = MISSING
    members: APINullable[List[GuildMember]] = MISSING
    member_count: APINullable[bool] = MISSING
    nsfw: APINullable[bool] = MISSING
    # Note: This is missing from discord's docs but in the api
    owner: APINullable[bool] = MISSING
    permissions: APINullable[str] = MISSING
    premium_subscription_count: APINullable[int] = MISSING
    presences: APINullable[List[PresenceUpdateEvent]] = MISSING
    stage_instances: APINullable[List[StageInstance]] = MISSING
    stickers: APINullable[List[Sticker]] = MISSING
    region: APINullable[Optional[str]] = MISSING
    threads: APINullable[List[Channel]] = MISSING
    # Guilds are considered available unless otherwise specified
    unavailable: APINullable[bool] = False
    voice_states: APINullable[List[VoiceState]] = MISSING
    widget_enabled: APINullable[bool] = MISSING
    widget_channel_id: APINullable[Optional[Snowflake]] = MISSING
    welcome_screen: APINullable[WelcomeScreen] = MISSING

    @classmethod
    async def from_id(cls, client: Client, _id: Union[int, Snowflake]) -> Guild:
        """
        Parameters
        ----------
        client : `~pincer.Client`
            Client object to use the http gateway from.
        _id : :class: `pincer.utils.snowflake.Snowflake`
            Guild ID.
        Returns
        -------
        :class: `~pincer.objects.guild.guild.Guild`
            The new guild object.
        """
        data = await client.http.get(f"/guilds/{_id}")
        channel_data = await client.http.get(f"/guilds/{_id}/channels")

        data["channels"]: List[Channel] = [
            Channel.from_dict({**i, "_client": client, "_http": client.http})
            for i in (channel_data or [])
        ]

        return Guild.from_dict(construct_client_dict(client, data))

    async def get_member(self, _id: int) -> GuildMember:
        """|coro|
        Fetches a GuildMember from its identifier
        Parameters
        ----------
        _id:
            The id of the guild member which should be fetched from the Discord
            gateway.
        Returns
        -------
        :class:`~pincer.objects.guild.member.GuildMember`
            A GuildMember object.
        """
        return await GuildMember.from_id(self._client, self.id, _id)

    @overload
    async def modify_member(
            self, *,
            _id: int,
            nick: Optional[str] = None,
            roles: Optional[List[Snowflake]] = None,
            mute: Optional[bool] = None,
            deaf: Optional[bool] = None,
            channel_id: Optional[Snowflake] = None
    ) -> GuildMember:
        """|coro|
        Modifies a member in the guild from its identifier and based on the
        keyword arguments provided.
        Parameters
        ----------
        _id : int
            Id of the member to modify
        nick : Optional[:class:`str`]
            New nickname for the member |default| :data:`None`
        roles : Optional[List[:class:`~pincer.utils.snowflake.Snowflake]]
            New roles for the member |default| :data:`None`
        mute : Optional[:class:`bool`]
            Whether the member is muted |default| :data:`None`
        deaf : Optional[:class:`bool`]
            Whether the member is deafened |default| :data:`None`
        channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake]
            Voice channel id to move to |default| :data:`None`
        Returns
        -------
        :class:`~pincer.objects.guild.member.GuildMember`
            The new member object.
        """
        ...

    async def modify_member(self, _id: int, **kwargs) -> GuildMember:
        data = await self._http.patch(
            f"guilds/{self.id}/members/{_id}",
            data=kwargs
        )
        return GuildMember.from_dict(construct_client_dict(self._client, data))

    async def ban(self, member_id: int, **kwargs):
        """|coro|
        Bans a guild member.
        Parameters
        ----------
        member_id : :class:`int`
            ID of the guild member to ban.
        \\*\\* kwargs
            Additional keyword arguments to ban the guild member with.
        """
        await self._http.put(f"guilds/{self.id}/bans/{member_id}", data=kwargs)

    async def kick(self, member_id: int):
        """|coro|
        Kicks a guild member.
        Parameters
        ----------
        member_id : :class:`int`
            ID of the guild member to kick.
        """
        await self._http.delete(f"guilds/{self.id}/members/{member_id}")

    @overload
    async def edit(
        self,
        *,
        name: Optional[str] = None,
        region: Optional[str] = None,
        verification_level: Optional[int] = None,
        default_message_notifications: Optional[int] = None,
        explicit_content_filter: Optional[int] = None,
        afk_channel_id: Optional[Snowflake] = None,
        afk_timeout: Optional[int] = None,
        icon: Optional[str] = None,
        owner_id: Optional[Snowflake] = None,
        splash: Optional[str] = None,
        discovery_splash: Optional[str] = None,
        banner: Optional[str] = None,
        system_channel_id: Optional[Snowflake] = None,
        system_channel_flags: Optional[int] = None,
        rules_channel_id: Optional[Snowflake] = None,
        public_updates_channel_id: Optional[Snowflake] = None,
        preferred_locale: Optional[str] = None,
        features: Optional[List[GuildFeature]] = None,
        description: Optional[str] = None
    ) -> Guild:
        """|coro|
        Modifies the guild

        Parameters
        ----------
        name : Optional[:class:`str`]
            Guild name |default| :data:`None`
        region : Optional[:class:`str`]
            Guild voice region ID |default| :data:`None`
        verification_level : Optional[:class:`int`]
            Verification level |default| :data:`None`
        default_message_notifications : Optional[:class:`int`]
            Default message notification level |default| :data:`None`
        explicit_content_filter : Optional[:class:`int`]
            Explicit content filter level |default| :data:`None`
        afk_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID for AFK channel |default| :data:`None`
        afk_timeout : Optional[:class:`int`]
            AFK timeout in seconds |default| :data:`None`
        icon : Optional[:class:`str`]
            base64 1024x1024 png/jpeg/gif image for the guild icon
            (can be animated gif when the server
            has the `ANIMATED_ICON` feature) |default| :data:`None`
        owner_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            User ID to transfer guild ownership to (must be owner) |default| :data:`None`
        splash : Optional[:class:`str`]
            base64 16:9 png/jpeg image for the guild splash (when the
            server has the `INVITE_SPLASH` feature) |default| :data:`None`
        discovery_splash : Optional[:class:`str`]
            base64 16:9 png/jpeg image for the guild discovery splash
            (when the server has the `DISCOVERABLE` feature) |default| :data:`None`
        banner : Optional[:class:`str`]
            base64 16:9 png/jpeg image for the guild banner (when the
            server has the `BANNER` feature) |default| :data:`None`
        system_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where guild notices such as welcome
            messages and boost events are posted |default| :data:`None`
        system_channel_flags : Optional[:class:`int`]
            System channel flags |default| :data:`None`
        rules_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where Community guilds display rules
            and/or guidelines |default| :data:`None`
        public_updates_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where admins and moderators of
            Community guilds receive notices from Discord |default| :data:`None`
        preferred_locale : Optional[:class:`str`]
            The preferred locale of a Community guild used in server
            discovery and notices from Discord; defaults to "en-US" |default| :data:`None`
        features : Optional[List[:class:`GuildFeature`]]
            Enabled guild features |default| :data:`None`
        description : Optional[:class:`str`]
            The description for the guild, if the guild is discoverable |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.Guild`
            The modified guild object.
        """
        ...

    async def edit(self, **kwargs) -> Guild:
        g = await self._http.patch(f"guilds/{self.id}", data=kwargs)
        return Guild.from_dict(construct_client_dict(self._client, g))

    async def preview(self) -> GuildPreview:
        """|coro|
        Previews the guild.
        Returns
        -------
        :class:`~pincer.objects.guild.guild.GuildPreview`
            The guild preview object.
        """
        data = await self._http.get(f"guilds/{self.id}/preview")
        return GuildPreview.from_dict(data)

    async def delete(self):
        """|coro|
        Deletes the guild. Returns `204 No Content` on success.
        """
        await self._http.delete(f"guilds/{self.id}")

    @classmethod
    def from_dict(cls, data) -> Guild:
        """
        Parameters
        ----------
        data : :class:`Dict`
            Guild data received from the discord API.
        Returns
        -------
        :class: `~pincer.objects.guild.guild.Guild`
            The new guild object.
        Raises
        :class: `~pincer.exceptions.UnavailableGuildError`
            The guild is unavailable due to a discord outage.
        """
        if data.get("unavailable", False):
            raise UnavailableGuildError(
                f"Guild \"{data['id']}\" is unavailable due to a discord"
                " outage."
            )

        return super().from_dict(data)


@dataclass
class UnavailableGuild(APIObject):
    id: Snowflake
    unavailable: bool = True



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild ban is removed."""

from ..core.dispatch import GatewayDispatch
from ..objects.events.guild import GuildBanRemoveEvent
from ..utils import Coro
from ..utils.conversion import construct_client_dict


async def guild_ban_remove_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_ban_remove`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild ban remove event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.guild.GuildBanRemoveEvent`]]
        ``on_guild_ban_remove_update`` and a ``GuildBanRemoveEvent``
    """

    return (
        "on_guild_ban_remove",
        [GuildBanRemoveEvent.from_dict(construct_client_dict(self, payload.data))],
    )


def export() -> Coro:
    return guild_ban_remove_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message is updated in a subscribed text channel"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import UserMessage
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..core.dispatch import GatewayDispatch


async def message_update_middleware(
    self,
    payload: GatewayDispatch
) -> Tuple[str, List[UserMessage]]:
    """|coro|


    Middleware for ``on_message_update`` event,
        generate a class for the message that has been updated.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the message update event event

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.message.user_message.UserMessage`]]
        ``on_message_update`` and a ``UserMessage``
    """
    return "on_message_update", [
        UserMessage.from_dict(construct_client_dict(self, payload.data))
    ]


def export():
    return message_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .channel import ChannelPinsUpdateEvent
from .error import DiscordError
from .gateway_commands import (
    Identify, Resume, RequestGuildMembers, UpdateVoiceState,
    StatusType, UpdatePresence
)
from .guild import (
    GuildBanAddEvent, GuildBanRemoveEvent, GuildEmojisUpdateEvent,
    GuildStickersUpdateEvent, GuildIntegrationsUpdateEvent,
    GuildMemberRemoveEvent, GuildMemberUpdateEvent, GuildMembersChunkEvent,
    GuildRoleCreateEvent, GuildRoleUpdateEvent, GuildRoleDeleteEvent
)
from .hello_ready import HelloEvent, ReadyEvent
from .integration import IntegrationDeleteEvent
from .invite import InviteCreateEvent, InviteDeleteEvent
from .message import (
    MessageDeleteEvent, MessageDeleteBulkEvent, MessageReactionAddEvent,
    MessageReactionRemoveEvent, MessageReactionRemoveAllEvent,
    MessageReactionRemoveEmojiEvent
)
from .presence import (
    ActivityType, ActivityTimestamp, ActivityEmoji, ActivityParty,
    ActivityAssets, ActivitySecrets, ActivityFlags, ActivityButton,
    Activity, ClientStatus, PresenceUpdateEvent
)
from .thread import ThreadListSyncEvent, ThreadMembersUpdateEvent
from .typing_start import TypingStartEvent
from .voice import VoiceServerUpdateEvent
from .webhook import WebhooksUpdateEvent

__all__ = (
    "Activity", "ActivityAssets", "ActivityButton", "ActivityEmoji",
    "ActivityFlags", "ActivityParty", "ActivitySecrets", "ActivityTimestamp",
    "ActivityType", "ChannelPinsUpdateEvent", "ClientStatus", "DiscordError",
    "GuildBanAddEvent", "GuildBanRemoveEvent", "GuildEmojisUpdateEvent",
    "GuildIntegrationsUpdateEvent", "GuildMemberRemoveEvent",
    "GuildMemberUpdateEvent", "GuildMembersChunkEvent", "GuildRoleCreateEvent",
    "GuildRoleDeleteEvent", "GuildRoleUpdateEvent", "GuildStickersUpdateEvent",
    "HelloEvent", "Identify", "IntegrationDeleteEvent", "InviteCreateEvent",
    "InviteDeleteEvent", "MessageDeleteBulkEvent", "MessageDeleteEvent",
    "MessageReactionAddEvent", "MessageReactionRemoveAllEvent",
    "MessageReactionRemoveEmojiEvent", "MessageReactionRemoveEvent",
    "PresenceUpdateEvent", "ReadyEvent", "RequestGuildMembers", "Resume",
    "StatusType", "ThreadListSyncEvent", "ThreadMembersUpdateEvent",
    "TypingStartEvent", "UpdatePresence", "UpdateVoiceState",
    "VoiceServerUpdateEvent", "WebhooksUpdateEvent"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import asyncio
import logging
from asyncio import TimerHandle, iscoroutinefunction
from datetime import timedelta
from typing import TYPE_CHECKING, Optional

from . import __package__
from .insertion import should_pass_cls
from ..exceptions import (
    TaskAlreadyRunning, TaskCancelError, TaskInvalidDelay,
    TaskIsNotCoroutine
)

if TYPE_CHECKING:
    from typing import Callable, Set
    from .types import Coro


_log = logging.getLogger(__package__)


class TaskScheduler:
    def __init__(self, client):
        """
        Used to create tasks
        """
        self.client = client
        self.tasks: Set[Task] = set()
        self._loop = asyncio.get_event_loop()

    def loop(
        self,
        days=0,
        weeks=0,
        hours=0,
        minutes=0,
        seconds=0,
        milliseconds=0,
        microseconds=0
    ) -> Callable[[Coro], Task]:
        """A decorator to create a task that repeat the given amount of t
        :Example usage:

        .. code-block:: python
            from pincer import Client
            from pincer.utils import TaskScheduler

            client = Client("token")
            task = TaskScheduler(client)

            @task.loop(minutes=3)
            async def my_task(self):
                ...

            my_task.start()
            client.run()

        Parameters
        ----------
        days : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        weeks : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        hours : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        minutes : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        seconds : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        milliseconds : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        microseconds : :class:`int`
            Days to wait between iterations.
            |default| ``0``
        Raises
        ------
        TaskIsNotCoroutine:
            The task is not a coroutine.
        TaskInvalidDelay:
            The delay is 0 or negative.
        """
        def decorator(func: Coro) -> Task:
            if not iscoroutinefunction(func):
                raise TaskIsNotCoroutine(
                    f'Task `{func.__name__}` is not a coroutine, '
                    'which is required for tasks.'
                )

            delay = timedelta(
                days=days,
                weeks=weeks,
                hours=hours,
                minutes=minutes,
                seconds=seconds,
                microseconds=microseconds,
                milliseconds=milliseconds
            ).total_seconds()

            if delay <= 0:
                raise TaskInvalidDelay(
                    f'Task `{func.__name__}` has a delay of {delay} seconds, '
                    'which is invalid. Delay must be greater than zero.'
                )

            return Task(self, func, delay)

        return decorator

    def register(self, task: Task):
        """Register a task.
        Parameters
        ----------
        task : :class:`~pincer.utils.tasks.Task`
            The task to register.
        """
        self.tasks.add(task)
        self.__execute(task)

    def __execute(self, task: Task):
        """Execute a task."""
        coro = task.coro(self.client) if task.client_required else task.coro()
        # Execute the coroutine
        asyncio.ensure_future(coro)

        # Schedule the coroutine's next execution
        task._handle = self._loop.call_later(task.delay, self.__execute, task)

    def close(self):
        """Gracefully stops any running task."""
        for task in self.tasks.copy():
            if task.running:
                task.cancel()


class Task:
    """A Task is a coroutine that is scheduled to repeat every x seconds.
    Use a TaskScheduler in order to create a task.
    Parameters
    ----------
    scheduler: :class:`~pincer.utils.tasks.TaskScheduler`
        The scheduler to use.
    coro: :class:`~pincer.utils.types.Coro`
        The coroutine to register as a task.
    delay: :class:`float`
        Delay between each iteration of the task.
    """

    def __init__(self, scheduler: TaskScheduler, coro: Coro, delay: float):
        self._scheduler = scheduler
        self.coro = coro
        self.delay = delay
        self._handle: Optional[TimerHandle] = None
        self._client_required = should_pass_cls(coro)

    def __del__(self):
        if self.running:
            self.cancel()
        else:
            # Did the user forgot to call task.start() ?
            _log.warning(
                "Task `%s` was not scheduled. Did you forget to start it ?",
                self.coro.__name__
            )

    @property
    def cancelled(self):
        """:class:`bool`: Check if the task has been cancelled or not."""
        return self.running and self._handle.cancelled()

    @property
    def running(self):
        """:class:`bool`: Check if the task is running."""
        return self._handle is not None

    def start(self):
        """Register the task in the TaskScheduler and start
        the execution of the task.
        """
        if self.running:
            raise TaskAlreadyRunning(
                f'Task `{self.coro.__name__}` is already running.', self
            )

        self._scheduler.register(self)

    def cancel(self):
        """Cancel the task."""
        if not self.running:
            raise TaskCancelError(
                f'Task `{self.coro.__name__}` is not running.', self
            )

        self._handle.cancel()
        if self in self._scheduler.tasks:
            self._scheduler.tasks.remove(self)

    @property
    def client_required(self):
        return self._client_required



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""Sent when a channel is created/joined on the client."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core.dispatch import GatewayDispatch
from ..objects.guild.channel import Channel
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..core.dispatch import GatewayDispatch


def channel_create_middleware(
        self,
        payload: GatewayDispatch
) -> Tuple[str, List[Channel]]:
    """|coro|

    Middleware for ``on_channel_creation`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.dispatch.GatewayDispatch`
        The data received from the ready event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.channel.Channel`]]
        ``"on_channel_creation"`` and a channel.
    """
    return "on_channel_creation", [
        Channel.from_dict(construct_client_dict(self, payload.data))
    ]


def export():
    return channel_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Dict

    from ..user.user import User
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


class CallbackType(IntEnum):
    """The types of response a client can give to a interaction.

    Attributes
    ----------
    PONG:
        ACK a Ping
    MESSAGE:
        Respond to an interaction with a message
    DEFERRED_MESSAGE:
        ACK an interaction and edit a response later, the user sees a loading
        state
    DEFERRED_UPDATE_MESSAGE:
        For components, ACK an interaction and edit the original message later
    UPDATE_MESSAGE:
        For components, edit the message the component was attached to
    """
    # noqa: E501
    PONG = 1
    MESSAGE = 4
    DEFERRED_MESSAGE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7


class InteractionType(IntEnum):
    """Represents the different types of interactions the client
    can have with a member.

    Attributes
    ----------
    PING:
        Ping an interaction.
    APPLICATION_COMMAND:
        A "slash" command.
    MESSAGE_COMPONENT:
        A ui component like buttons and selects.
    """
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3


@dataclass
class MessageInteraction(APIObject):
    """Represents a Discord Message Interaction object

    This is sent on the message object when the message
    is a response to an Interaction without an existing message.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the interaction
    type: :class:`~pincer.objects.app.interaction_base.InteractionType`
        The type of interaction
    name: :class:`str`
        The name of the application command
    user: :class:`~pincer.objects.user.user.User`
        The user who invoked the interaction
    member: APINullable[Dict]
        The member who invoked the interaction
    """
    id: Snowflake
    type: InteractionType
    name: str
    user: User

    member: APINullable[Dict] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when a user's voice state changes in a subscribed voice channel
(mute, volume, etc.)
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.user.voice_state import VoiceState
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..core.dispatch import GatewayDispatch


async def voice_state_update_middleware(
    self,
    payload: GatewayDispatch
) -> Tuple[str, List[VoiceState]]:
    """|coro|
    Middleware for ``on_voice_state_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the voice state update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.user.voice_state.VoiceState`]]
        ``on_voice_state_update`` and a ``VoiceState``
    """
    # noqa: E501

    return "on_voice_state_update", [
        VoiceState.from_dict(construct_client_dict(self, payload.data))
    ]


def export():
    return voice_state_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

import logging
from glob import glob
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING

from ..utils.directory import chdir

if TYPE_CHECKING:
    from typing import Dict
    from ..utils.types import Coro


_log = logging.getLogger(__package__)


def get_middleware() -> Dict[str, Coro]:
    middleware_list: Dict[str, Coro] = {}

    with chdir(Path(__file__).parent.resolve()):

        for middleware_path in glob("*.py"):
            if middleware_path.startswith("__"):
                continue

            event = middleware_path[:-3]

            try:
                middleware_list[event] = getattr(
                    import_module(f".{event}", package=__name__),
                    "export"
                )()
            except AttributeError:
                _log.warning(
                    f"Middleware {middleware_path} expected an `export` method."
                )

                continue

                # raise NoExportMethod(
                #    f"Middleware module `{middleware_path}` expected an "
                #    "`export` method but none was found!"
                # )

    return middleware_list


middleware: Dict[str, Coro] = get_middleware()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING, APINullable

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..user.user import User
    from ..guild.guild import Guild
    from ..app.application import Application


@dataclass
class HelloEvent(APIObject):
    """Sent on connection to the websocket.
    Defines the heartbeat interval that the client should heartbeat to.

    Attributes
    ----------
    heartbeat_interval: :class:`int`
        The interval (in milliseconds) the client should heartbeat with
    """
    heartbeat_interval: int


@dataclass
class ReadyEvent(APIObject):
    """Dispatched when a client has completed the initial
    handshake with the gateway (for new sessions).

    Attributes
    ----------
    v: :class:`int`
        Gateway version
    user: :class:`~pincer.objects.user.user.User`
        Information about the user including email
    guilds: List[:class:`~pincer.objects.guild.guild.Guild`]
        The guilds the user is in
    session_id: :class:`str`
        Used for resuming connections
    application: :class:`~pincer.objects.app.application.Application`
        Contains ``id`` and ``flags``
    shard: APINullable[Tuple[:class:`int`, :class:`int`]]
        The shard information associated
        with this session, if sent when identifying
    """
    v: int
    user: User
    guilds: List[Guild]
    session_id: str
    application: Application

    shard: APINullable[Tuple[int, int]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from typing import Optional

    from ...utils.snowflake import Snowflake


@dataclass
class GuildWidget(APIObject):
    """Represents a Discord Guild Widget object

    Attributes
    ----------
    enabled: :class:`bool`
        Whether the widget is enabled
    channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The widget channel id
    """
    enabled: bool
    channel_id: Optional[Snowflake]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .dispatch import GatewayDispatch
from .gateway import Dispatcher
from .heartbeat import Heartbeat
from .http import HTTPClient


__all__ = (
    "Dispatcher", "GatewayDispatch", "HTTPClient", "Heartbeat"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
import re
from asyncio import iscoroutinefunction, gather
from copy import deepcopy
from functools import partial
from inspect import Signature, isasyncgenfunction
from typing import TYPE_CHECKING, get_origin, get_args, Union, Tuple, List

from . import __package__
from .utils.snowflake import Snowflake
from .exceptions import (
    CommandIsNotCoroutine,
    CommandAlreadyRegistered,
    TooManyArguments,
    InvalidArgumentAnnotation,
    CommandDescriptionTooLong,
    InvalidCommandGuild,
    InvalidCommandName,
    ForbiddenError,
)
from .objects import (
    ThrottleScope,
    AppCommand,
    Role,
    User,
    Channel,
    Guild,
    MessageContext,
)
from .objects.app import (
    AppCommandOptionType,
    AppCommandOption,
    AppCommandOptionChoice,
    ClientCommandStructure,
    AppCommandType,
)
from .utils import get_index, should_pass_ctx
from .utils.signature import get_signature_and_params
from .utils.types import Coro, MISSING, choice_value_types, Choices
from .utils.types import Singleton, TypeCache, Descripted

if TYPE_CHECKING:
    from typing import Any, Optional, Dict

COMMAND_NAME_REGEX = re.compile(r"^[\w-]{1,32}$")

_log = logging.getLogger(__package__)

_options_type_link = {
    # TODO: Implement other types:
    Signature.empty: AppCommandOptionType.STRING,
    str: AppCommandOptionType.STRING,
    int: AppCommandOptionType.INTEGER,
    bool: AppCommandOptionType.BOOLEAN,
    float: AppCommandOptionType.NUMBER,
    User: AppCommandOptionType.USER,
    Channel: AppCommandOptionType.CHANNEL,
    Role: AppCommandOptionType.ROLE,
}

if TYPE_CHECKING:
    from .client import Client


def command(
    func=None,
    *,
    name: Optional[str] = None,
    description: Optional[str] = "Description not set",
    enable_default: Optional[bool] = True,
    guild: Union[Snowflake, int, str] = None,
    cooldown: Optional[int] = 0,
    cooldown_scale: Optional[float] = 60,
    cooldown_scope: Optional[ThrottleScope] = ThrottleScope.USER,
):
    """A decorator to create a command to register and respond to
    with the discord API from a function.

    str - String
    int - Integer
    bool - Boolean
    float - Number
    pincer.objects.User - User
    pincer.objects.Channel - Channel
    pincer.objects.Role - Role
    Mentionable is not implemented

    .. code-block:: python3

        class Bot(Client):
            @command(
                name="test",
                description="placeholder"
            )
            async def test_command(
                self,
                ctx,
                amount: int,
                name: Descripted[str, "ah yes"],
                letter: Choices["a", "b", "c"]
            ):
                return Message(
                    f"You chose {amount}, {name}, {letter}",
                    flags=InteractionFlags.EPHEMERAL
                )

    References from above:
        :class:`~client.Client`,
        :class:`~objects.message.message.Message`,
        :class:`~utils.types.Choices`,
        :class:`~utils.types.Descripted`,
        :class:`~objects.app.interactions.InteractionFlags`

    Parameters
    ----------
    name : Optional[:class:`str`]
        The name of the command |default| :data:`None`
    description : Optional[:class:`str`]
        The description of the command |default| ``Description not set``
    enable_default : Optional[:class:`bool`]
        Whether the command is enabled by default |default| :data:`True`
    guild : Optional[Union[:class:`~pincer.utils.snowflake.Snowflake`, :class:`int`, :class:`str`]]
        What guild to add it to (don't specify for global) |default| :data:`None`
    cooldown : Optional[:class:`int`]
        The amount of times in the cooldown_scale the command can be invoked
        |default| ``0``
    cooldown_scale : Optional[:class:`float`]
        The 'checking time' of the cooldown |default| ``60``
    cooldown_scope : :class:`~pincer.objects.app.throttle_scope.ThrottleScope`
        What type of cooldown strategy to use |default| :attr:`ThrottleScope.USER`

    Raises
    ------
    CommandIsNotCoroutine
        If the command function is not a coro
    InvalidCommandName
        If the command name does not follow the regex ``^[\\w-]{1,32}$``
    InvalidCommandGuild
        If the guild id is invalid
    CommandDescriptionTooLong
        Descriptions max 100 characters
        If the annotation on an argument is too long (also max 100)
    CommandAlreadyRegistered
        If the command already exists
    TooManyArguments
        Max 25 arguments to pass for commands
    InvalidArgumentAnnotation
        Annotation amount is max 25,
        Not a valid argument type,
        Annotations must consist of name and value
    """
    # noqa: E501
    
    if func is None:
        return partial(
            command,
            name=name,
            description=description,
            enable_default=enable_default,
            guild=guild,
            cooldown=cooldown,
            cooldown_scale=cooldown_scale,
            cooldown_scope=cooldown_scope,
        )

    if not iscoroutinefunction(func) and not isasyncgenfunction(func):
        raise CommandIsNotCoroutine(
            f"Command with call `{func.__name__}` is not a coroutine, "
            "which is required for commands."
        )

    cmd = name or func.__name__

    if not re.match(COMMAND_NAME_REGEX, cmd):
        raise InvalidCommandName(
            f"Command `{cmd}` doesn't follow the name requirements."
            "Ensure to match the following regex:"
            f" {COMMAND_NAME_REGEX.pattern}"
        )

    try:
        guild_id = Snowflake(guild) if guild else MISSING
    except ValueError:
        raise InvalidCommandGuild(
            f"Command with call `{func.__name__}` its `guilds` parameter "
            "contains a non valid guild id."
        )

    if len(description) > 100:
        raise CommandDescriptionTooLong(
            f"Command `{cmd}` (`{func.__name__}`) its description exceeds "
            "the 100 character limit."
        )

    if reg := ChatCommandHandler.register.get(cmd):
        raise CommandAlreadyRegistered(
            f"Command `{cmd}` (`{func.__name__}`) has already been "
            f"registered by `{reg.call.__name__}`."
        )

    sig, params = get_signature_and_params(func)
    pass_context = should_pass_ctx(sig, params)

    if len(params) > (25 + pass_context):
        raise TooManyArguments(
            f"Command `{cmd}` (`{func.__name__}`) can only have 25 "
            f"arguments (excluding the context and self) yet {len(params)} "
            "were provided!"
        )

    options: List[AppCommandOption] = []

    for idx, param in enumerate(params):
        if idx == 0 and pass_context:
            continue

        annotation, required = sig[param].annotation, True

        # ctx is type MessageContext but should not be included in the
        # slash command
        if annotation == MessageContext:
            return

        argument_description: Optional[str] = None
        choices: List[AppCommandOptionChoice] = []

        if isinstance(annotation, str):
            TypeCache()
            annotation = eval(annotation, TypeCache.cache, globals())

        if isinstance(annotation, Descripted):
            argument_description = annotation.description
            annotation = annotation.key

            if len(argument_description) > 100:
                raise CommandDescriptionTooLong(
                    f"Tuple annotation `{annotation}` on parameter "
                    f"`{param}` in command `{cmd}` (`{func.__name__}`), "
                    "argument description too long. (maximum length is 100 "
                    "characters)"
                )

        if get_origin(annotation) is Union:
            args = get_args(annotation)
            if type(None) in args:
                required = False

            # Do NOT use isinstance as this is a comparison between
            # two values of the type type and isinstance does NOT
            # work here.
            union_args = [t for t in args if t is not type(None)]

            annotation = (
                get_index(union_args, 0)
                if len(union_args) == 1
                else Tuple[List]
            )

        if get_origin(annotation) is Choices:
            args = get_args(annotation)

            if len(args) > 25:
                raise InvalidArgumentAnnotation(
                    f"Choices/Literal annotation `{annotation}` on "
                    f"parameter `{param}` in command `{cmd}` "
                    f"(`{func.__name__}`) amount exceeds limit of 25 items!"
                )

            choice_type = type(args[0])

            if choice_type is Descripted:
                choice_type = type(args[0].key)

            for choice in args:
                choice_description = choice
                if isinstance(choice, Descripted):
                    choice_description = choice.description
                    choice = choice.key

                    if choice_type is tuple:
                        choice_type = type(choice)

                if type(choice) not in choice_value_types:
                    # Properly get all the names of the types
                    valid_types = list(
                        map(lambda x: x.__name__, choice_value_types)
                    )
                    raise InvalidArgumentAnnotation(
                        f"Choices/Literal annotation `{annotation}` on "
                        f"parameter `{param}` in command `{cmd}` "
                        f"(`{func.__name__}`), invalid type received. "
                        "Value must be a member of "
                        f"{', '.join(valid_types)} but "
                        f"{type(choice).__name__} was given!"
                    )
                elif not isinstance(choice, choice_type):
                    raise InvalidArgumentAnnotation(
                        f"Choices/Literal annotation `{annotation}` on "
                        f"parameter `{param}` in command `{cmd}` "
                        f"(`{func.__name__}`), all values must be of the "
                        "same type!"
                    )

                choices.append(
                    AppCommandOptionChoice(
                        name=choice_description, value=choice
                    )
                )

            annotation = choice_type

        param_type = _options_type_link.get(annotation)

        if not param_type:
            raise InvalidArgumentAnnotation(
                f"Annotation `{annotation}` on parameter "
                f"`{param}` in command `{cmd}` (`{func.__name__}`) is not "
                "a valid type."
            )

        options.append(
            AppCommandOption(
                type=param_type,
                name=param,
                description=argument_description or "Description not set",
                required=required,
                choices=choices or MISSING,
            )
        )

    ChatCommandHandler.register[cmd] = ClientCommandStructure(
        call=func,
        cooldown=cooldown,
        cooldown_scale=cooldown_scale,
        cooldown_scope=cooldown_scope,
        app=AppCommand(
            name=cmd,
            description=description,
            type=AppCommandType.CHAT_INPUT,
            default_permission=enable_default,
            options=options,
            guild_id=guild_id,
        ),
    )

    _log.info(f"Registered command `{cmd}` to `{func.__name__}`.")
    return func


class ChatCommandHandler(metaclass=Singleton):
    """Metaclass containing methods used to handle various commands

    Attributes
    ----------
    client: :class:`Client`
        The client object
    managers: Dict
        Dictionary of managers
    register: Dict
        Dictionary of :class:`~objects.app.command.ClientCommandStructure`
    """

    managers: Dict[str, Any] = {}
    register: Dict[str, ClientCommandStructure] = {}

    # Endpoints:
    __get = "/commands"
    __delete = "/commands/{command.id}"
    __update = "/commands/{command.id}"
    __add = "/commands"
    __add_guild = "/guilds/{command.guild_id}/commands"
    __get_guild = "/guilds/{guild_id}/commands"
    __update_guild = "/guilds/{command.guild_id}/commands/{command.id}"
    __delete_guild = "/guilds/{command.guild_id}/commands/{command.id}"

    def __init__(self, client: Client):
        self.client = client
        self._api_commands: List[AppCommand] = []
        logging.debug(
            "%i commands registered.", len(ChatCommandHandler.register.items())
        )
        self.client.throttler.throttle = dict(
            map(
                lambda cmd: (cmd.call, {}), ChatCommandHandler.register.values()
            )
        )

        self.__prefix = f"applications/{self.client.bot.id}"

    async def get_commands(self) -> List[AppCommand]:
        """|coro|

        Get a list of app commands

        Returns
        -------
        List[:class:`~pincer.objects.app.command.AppCommand`]
            List of commands.
        """
        # TODO: Update if discord adds bulk get guild commands
        guild_commands = await gather(
            *map(
                lambda guild: self.client.http.get(
                    self.__prefix
                    + self.__get_guild.format(
                        guild_id=guild.id if isinstance(guild, Guild) else guild
                    )
                ),
                self.client.guilds,
            )
        )
        return list(
            map(
                AppCommand.from_dict,
                await self.client.http.get(self.__prefix + self.__get)
                + [cmd for guild in guild_commands for cmd in guild],
            )
        )

    async def remove_command(self, cmd: AppCommand, keep=False):
        """|coro|

        Remove a specific command

        Parameters
        ----------
        cmd : :class:`~pincer.objects.app.command.AppCommand`
            What command to delete
        keep : bool
            Whether the command should be removed from the ChatCommandHandler.
            Set to :data:`True` to keep the command.
            |default| :data:`False`
        """
        # TODO: Update if discord adds bulk delete commands
        remove_endpoint = self.__delete_guild if cmd.guild_id else self.__delete

        await self.client.http.delete(
            self.__prefix + remove_endpoint.format(command=cmd)
        )

        if not keep and ChatCommandHandler.register.get(cmd.name):
            del ChatCommandHandler.register[cmd.name]

    async def remove_commands(
        self, commands: List[AppCommand], /, keep: List[AppCommand] = None
    ):
        """|coro|

        Remove a list of commands

        Parameters
        ----------
        commands : List[:class:`~pincer.objects.app.command.AppCommand`]
            List of commands to delete
        keep: List[:class:`~pincer.objects.app.command.AppCommand`]
            List of commands that should not be removed from the
            ChatCommandHandler.
            |default| :data:`None`
        """
        await gather(
            *list(
                map(
                    lambda cmd: self.remove_command(cmd, cmd in (keep or [])),
                    commands,
                )
            )
        )

    async def update_command(self, cmd: AppCommand, changes: Dict[str, Any]):
        """|coro|

        Update a command with changes

        Parameters
        ----------
        cmd : :class:`~objects.app.command.AppCommand`
            What command to update
        changes : Dict[:class:`str`, Any]
            Dictionary of changes
        """
        # TODO: Update if discord adds bulk update commands
        update_endpoint = self.__update_guild if cmd.guild_id else self.__update

        await self.client.http.patch(
            self.__prefix + update_endpoint.format(command=cmd), data=changes
        )

        for key, value in changes.items():
            setattr(ChatCommandHandler.register[cmd.name], key, value)

    async def update_commands(
        self, to_update: Dict[AppCommand, Dict[str, Any]]
    ):
        """|coro|

        Update a list of app commands with changes

        Parameters
        ----------
        to_update : Dict[:class:`~objects.app.command.AppCommand`, Dict[:class:`str`, Any]]
            Dictionary of commands to changes where changes is a dictionary too
        """  # noqa: E501
        await gather(
            *list(
                map(
                    lambda cmd: self.update_command(cmd[0], cmd[1]),
                    to_update.items(),
                )
            )
        )

    async def add_command(self, cmd: AppCommand):
        """|coro|

        Add an app command

        Parameters
        ----------
        cmd : :class:`~pincer.objects.app.command.AppCommand`
            Command to add
        """
        add_endpoint = self.__add

        if cmd.guild_id:
            add_endpoint = self.__add_guild.format(command=cmd)

        res = await self.client.http.post(
            self.__prefix + add_endpoint, data=cmd.to_dict()
        )

        ChatCommandHandler.register[cmd.name].app.id = Snowflake(res["id"])

    async def add_commands(self, commands: List[AppCommand]):
        """|coro|

        Add a list of app commands

        Parameters
        ----------
        commands : List[:class:`~pincer.objects.app.command.AppCommand`]
            List of command objects to add
        """
        await gather(*list(map(lambda cmd: self.add_command(cmd), commands)))

    async def __init_existing_commands(self):
        """|coro|

        Initiate existing commands
        """
        try:
            self._api_commands = await self.get_commands()

        except ForbiddenError:
            logging.error("Cannot retrieve slash commands, skipping...")

            return

        for api_cmd in self._api_commands:
            cmd = ChatCommandHandler.register.get(api_cmd.name)
            if cmd and cmd.app == api_cmd:
                cmd.app = api_cmd

    async def __remove_unused_commands(self):
        """|coro|

        Remove commands that are registered by discord but not in use
        by the current client
        """
        registered_commands = list(
            map(
                lambda registered_cmd: registered_cmd.app,
                ChatCommandHandler.register.values(),
            )
        )
        keep = []

        def predicate(target: AppCommand) -> bool:
            for reg_cmd in registered_commands:
                reg_cmd: AppCommand = reg_cmd
                if target == reg_cmd:
                    return False
                elif target.name == reg_cmd.name:
                    keep.append(target)
            return True

        to_remove = list(filter(predicate, self._api_commands))

        await self.remove_commands(to_remove, keep=keep)

        self._api_commands = list(
            filter(lambda cmd: cmd not in to_remove, self._api_commands)
        )

    async def __update_existing_commands(self):
        """|coro|

        Update all commands where its structure doesn't match the
        structure that discord has registered.
        """
        to_update: Dict[AppCommand, Dict[str, Any]] = {}

        def get_changes(api: AppCommand, local: AppCommand) -> Dict[str, Any]:
            update: Dict[str, Any] = {}

            if api.description != local.description:
                update["description"] = local.description

            if api.default_permission != local.default_permission:
                update["default_permission"] = local.default_permission

            options: List[Dict[str, Any]] = []
            if api.options is not MISSING:
                if len(api.options) == len(local.options):

                    def get_option(
                        args: Tuple[int, Any]
                    ) -> Optional[Dict[str, Any]]:
                        index, api_option = args

                        if opt := get_index(local.options, index):
                            return opt.to_dict()

                    options = list(
                        filter(
                            lambda opt: opt is not None,
                            map(get_option, enumerate(api.options)),
                        )
                    )
                else:
                    options = local.options

            if (
                api.options is not MISSING
                and list(map(AppCommandOption.from_dict, options))
                != api.options
            ):
                update["options"] = options

            return update

        for idx, api_cmd in enumerate(self._api_commands):
            for loc_cmd in ChatCommandHandler.register.values():
                if api_cmd.name != loc_cmd.app.name:
                    continue

                changes = get_changes(api_cmd, loc_cmd.app)

                if not changes:
                    continue

                api_update = []
                if changes.get("options"):
                    for option in changes["options"]:
                        api_update.append(
                            option.to_dict()
                            if isinstance(option, AppCommandOption)
                            else option
                        )

                to_update[api_cmd] = {"options": api_update}

                for key, change in changes.items():
                    if key == "options":
                        self._api_commands[idx].options = list(
                            map(AppCommandOption.from_dict, change)
                        )
                    else:
                        setattr(self._api_commands[idx], key, change)

        await self.update_commands(to_update)

    async def __add_commands(self):
        """|coro|

        Add all new commands which have been registered by the decorator
        to Discord
        """
        to_add = deepcopy(ChatCommandHandler.register)
        for reg_cmd in self._api_commands:
            try:
                del to_add[reg_cmd.name]
            except IndexError:
                pass

        await self.add_commands(list(map(lambda cmd: cmd.app, to_add.values())))

    async def initialize(self):
        """|coro|

        Call methods of this class to refresh all app commands
        """
        await self.__init_existing_commands()
        await self.__remove_unused_commands()
        await self.__update_existing_commands()
        await self.__add_commands()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

import logging
from asyncio import sleep
from json import dumps
from typing import Protocol, TYPE_CHECKING

from aiohttp import ClientSession, ClientResponse

from . import __package__
from .._config import GatewayConfig
from ..exceptions import (
    NotFoundError, BadRequestError, NotModifiedError, UnauthorizedError,
    ForbiddenError, MethodNotAllowedError, RateLimitError, ServerError,
    HTTPError
)

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union

    from aiohttp.client import _RequestContextManager
    from aiohttp.payload import Payload
    from aiohttp.typedefs import StrOrURL


_log = logging.getLogger(__package__)


class HttpCallable(Protocol):
    """Aiohttp HTTP method."""
    __name__: str

    def __call__(
            self, url: StrOrURL, *,
            allow_redirects: bool = True,
            method: Optional[Union[Dict, str, Payload]] = None,
            **kwargs: Any
    ) -> _RequestContextManager:
        ...


class HTTPClient:
    """Interacts with Discord API through HTTP protocol

    Attributes
    ----------
    url: :class:`str`
        ``f"https://discord.com/api/v{version}"`` "Base url for all HTTP requests"
    max_tts: :class:`int`
        Max amount of attempts after error code 5xx
    """

    def __init__(self, token: str, *, version: int = None, ttl: int = 5):
        """
        Instantiate a new HttpApi object.

        token:
            Discord API token

        Keyword Arguments:

        version:
            The discord API version.
            See `<https://discord.com/developers/docs/reference#api-versioning>`_.
        ttl:
            Max amount of attempts after error code 5xx
        """
        version = version or GatewayConfig.version
        self.url: str = f"https://discord.com/api/v{version}"
        self.max_ttl: int = ttl

        headers: Dict[str, str] = {
            "Authorization": f"Bot {token}",
        }
        self.__session: ClientSession = ClientSession(headers=headers)

        self.__http_exceptions: Dict[int, HTTPError] = {
            304: NotModifiedError(),
            400: BadRequestError(),
            401: UnauthorizedError(),
            403: ForbiddenError(),
            404: NotFoundError(),
            405: MethodNotAllowedError(),
            429: RateLimitError()
        }

    # for with block
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def close(self):
        """|coro|

        Closes the aiohttp session
        """
        await self.__session.close()

    async def __send(
            self,
            method: HttpCallable,
            endpoint: str, *,
            content_type: str = "application/json",
            data: Optional[Union[Dict, str, Payload]] = None,
            headers: Optional[Dict[str, Any]] = None,
            __ttl: int = None
    ) -> Optional[Dict]:
        """
        Send an api request to the Discord REST API.

        method:
            The method for the request. (eg GET or POST)

        endpoint:
            The endpoint to which the request will be sent.

        Keyword Arguments:

        content_type:
            The request's content type.

        data:
            The data which will be added to the request.

        __ttl:
            Private param used for recursively setting the retry amount.
            (Eg set to 1 for 1 max retry)
        """
        ttl = __ttl or self.max_ttl

        if ttl == 0:
            logging.error(
                # TODO: print better method name
                f"{method.__name__.upper()} {endpoint} has reached the "
                f"maximum retry count of {self.max_ttl}."
            )

            raise ServerError(f"Maximum amount of retries for `{endpoint}`.")

        if isinstance(data, dict):
            data = dumps(data)

        # TODO: print better method name
        # TODO: Adjust to work non-json types
        _log.debug(f"{method.__name__.upper()} {endpoint} | {data}")

        url = f"{self.url}/{endpoint}"
        async with method(
                url,
                data=data,
                headers={
                    "Content-Type": content_type,
                    **(headers or {})
                }
        ) as res:
            return await self.__handle_response(
                res, method, endpoint, content_type, data, ttl
            )

    async def __handle_response(
            self,
            res: ClientResponse,
            method: HttpCallable,
            endpoint: str,
            content_type: str,
            data: Optional[str],
            __ttl: int,
    ) -> Optional[Dict]:
        """
        Handle responses from the discord API.

        Side effects:
            If a 5xx error code is returned it will retry the request.

        res:
            The response from the discord API.

        method:
            The method which was used to call the endpoint.

        endpoint:
            The endpoint to which the request was sent.

        content_type:
            The request's content type.

        data:
            The data which was added to the request.

        __ttl:
            Private param used for recursively setting the retry amount.
            (Eg set to 1 for 1 max retry)
        """
        _log.debug(f"Received response for {endpoint} | {await res.text()}")
        if res.ok:
            if res.status == 204:
                _log.debug(
                    "Request has been sent successfully. "
                )
                return

            _log.debug(
                "Request has been sent successfully. "
                "Returning json response."
            )

            return await res.json()

        exception = self.__http_exceptions.get(res.status)

        if exception:
            if isinstance(exception, RateLimitError):
                timeout = (await res.json()).get("retry_after", 40)

                _log.exception(
                    f"RateLimitError: {res.reason}."
                    f" Retrying in {timeout} seconds"
                )
                await sleep(timeout)
                return await self.__send(
                    method,
                    endpoint,
                    content_type=content_type,
                    data=data
                )

            _log.error(
                f"An http exception occurred while trying to send "
                f"a request to {endpoint}. ({res.status}, {res.reason})"
            )

            exception.__init__(res.reason)
            raise exception

        # status code is guaranteed to be 5xx
        retry_in = 1 + (self.max_ttl - __ttl) * 2

        _log.debug(
            "Server side error occurred with status code "
            f"{res.status}. Retrying in {retry_in}s."
        )

        await sleep(retry_in)

        # try sending it again
        return await self.__send(
            method,
            endpoint,
            content_type=content_type,
            __ttl=__ttl - 1,
            data=data
        )

    async def delete(
            self,
            route: str,
            headers: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict]:
        """|coro|

        Sends a delete request to a Discord REST endpoint.

        Parameters
        ----------
        route : :class:`str`
            The Discord REST endpoint to send a delete request to.
        headers: Optional[Dict[:class:`str`, Any]]
            The request headers.
            |default| ``None``

        Returns
        -------
        Optional[Dict]
            The response from discord.
        """
        return await self.__send(
            self.__session.delete,
            route,
            headers=headers
        )

    async def get(self, route: str) -> Optional[Dict]:
        """|coro|

        Sends a get request to a Discord REST endpoint.

        Parameters
        ----------
        route : :class:`str`
            The Discord REST endpoint to send a get request to.

        Returns
        -------
        Optional[Dict]
            The response from discord.
        """
        return await self.__send(self.__session.get, route)

    async def head(self, route: str) -> Optional[Dict]:
        """|coro|

        Sends a head request to a Discord REST endpoint.

        Parameters
        ----------
        route : :class:`str`
            The Discord REST endpoint to send a head request to.

        Returns
        -------
        Optional[Dict]
            The response from discord.
        """
        return await self.__send(self.__session.head, route)

    async def options(self, route: str) -> Optional[Dict]:
        """|coro|

        Sends a options request to a Discord REST endpoint.

        Parameters
        ----------
        route : :class:`str`
            The Discord REST endpoint to send a options request to.

        Returns
        -------
        Optional[Dict]
            The response from discord.
        """
        return await self.__send(self.__session.options, route)

    async def patch(
            self,
            route: str,
            data: Optional[Dict] = None,
            content_type: str = "application/json",
            headers: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict]:
        """|coro|

        Sends a patch request to a Discord REST endpoint.

        Parameters
        ----------
        route : :class:`str``
            The Discord REST endpoint to send a patch request to.
        data : Dict
            The update data for the patch request.
        content_type: :class:`str`
            Body content type.
            |default| ``application/json``
        headers: Optional[Dict[:class:`str`, Any]]
            The request headers.

        Returns
        -------
        Optional[Dict]
            JSON response from the discord API.
        """
        return await self.__send(
            self.__session.patch,
            route,
            content_type=content_type,
            data=data,
            headers=headers
        )

    async def post(
            self,
            route: str,
            data: Optional[Dict] = None,
            content_type: str = "application/json",
            headers: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict]:
        """|coro|

        Sends a post request to a Discord REST endpoint

        Parameters
        ----------
        route : :class:`str`
            The Discord REST endpoint to send a patch request to.
        data : Dict
            The update data for the patch request.
        content_type : :class:`str`
            Body content type. |default| ``application/json``
        headers: Optional[Dict[:class:`str`, Any]]
            The request headers.

        Returns
        -------
        Optional[Dict]
            JSON response from the discord API.
        """
        return await self.__send(
            self.__session.post,
            route,
            content_type=content_type,
            data=data,
            headers=headers
        )

    async def put(
            self,
            route: str,
            data: Optional[Dict] = None,
            content_type: str = "application/json",
            headers: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict]:
        """|coro|

        Sends a put request to a Discord REST endpoint

        Parameters
        ----------
        route : :class:`str`
            The Discord REST endpoint to send a patch request to.
        data : Dict
            The update data for the patch request.
        content_type : :class:`str`
            Body content type. |default| ``application/json``
        headers: Optional[Dict[:class:`str`, Any]]
            The request headers.

        Returns
        -------
        Optional[Dict]
            JSON response from the discord API.
        """
        return await self.__send(
            self.__session.put,
            route,
            content_type=content_type,
            data=data,
            headers=headers
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from ...utils.snowflake import Snowflake


@dataclass
class FollowedChannel(APIObject):
    """Represents a Discord Followed Channel object

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        Source channel id
    webhook_id: :class:`~pincer.utils.snowflake.Snowflake`
        Created target webhook id
    """
    channel_id: Snowflake
    webhook_id: Snowflake



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when properties about a user changes"""

from ..core.dispatch import GatewayDispatch
from ..objects.user import User
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def user_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_user_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the user update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.user.user.User`]]
        ``on_user_update`` and a ``User``
    """
    return "on_user_update", [
        User.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return user_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from inspect import signature, isclass
from typing import Callable

from .insertion import should_pass_cls


def get_signature_and_params(func: Callable):
    """Get the parameters and signature from a coroutine.

    func: Callable
        The coroutine from whom the information should be extracted.

    Returns
    -------
    Tuple[List[Union[:class:`str`, :class:`inspect.Parameter`]]]
        Signature and ist of parameters of the coroutine.
    """
    if isclass(func):
        func = getattr(func, "__init__")

        if func is object.__init__:
            return [], []

    sig = signature(func).parameters
    params = list(sig)

    if should_pass_cls(func):
        del params[0]

    return sig, params


def get_params(func: Callable):
    """Get the parameters from a coroutine.

    func: Callable
        The coroutine from whom the information should be extracted.

    Returns
    -------
    List[Union[:class:`str`, :class:`inspect.Parameter`]]
        List of parameters of the coroutine.
    """
    return get_signature_and_params(func)[1]



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils import APIObject

if TYPE_CHECKING:
    from ...utils.snowflake import Snowflake


@dataclass
class Overwrite(APIObject):
    """Represents a Discord Overwrite object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Role or user id
    type: :class:`int`
        Either 0 (role) or 1 (member)
    allow: :class:`str`
        Permission bit set
    deny: :class:`str`
        Permission bit set
    """
    id: Snowflake
    type: int
    allow: str
    deny: str



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a user adds a reaction to a message"""

from ..core.dispatch import GatewayDispatch
from ..objects import GuildMember, Emoji
from ..objects.events.message import MessageReactionAddEvent
from ..utils.conversion import construct_client_dict


async def message_reaction_add_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_message_reaction_add`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the message reaction add event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.message.MessageReactionAddEvent`]]
        ``on_message_reaction_add`` and an ``MessageReactionAddEvent``
    """

    return "on_message_reaction_add", [
        MessageReactionAddEvent.from_dict(
            construct_client_dict(
                self,
                {
                    "member": GuildMember.from_dict(
                        construct_client_dict(self, payload.data.pop("member"))
                    ),
                    "emoji": Emoji.from_dict(
                        construct_client_dict(self, payload.data.pop("emoji"))
                    ),
                    **payload.data
                }
            ))
    ]


def export():
    return message_reaction_add_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations


class Snowflake(int):
    """Discord utilizes Twitter's snowflake format for uniquely
    identifiable descriptors (IDs).

    These IDs are guaranteed to be unique across all of Discord,
    except in some unique scenarios in which child objects
    share their parent's ID.

    Because Snowflake IDs are up to 64 bits in size (e.g. a uint64),
    they are always returned as strings in the HTTP API
    to prevent integer overflows in some languages.
    """
    _MAX_VALUE: int = 9223372036854775807
    _MIN_VALUE: int = 0

    def __init__(self, _):
        super().__init__()

        if self < self._MIN_VALUE:
            raise ValueError(
                "snowflake value should be greater than or equal to 0."
            )

        if self > self._MAX_VALUE:
            raise ValueError(
                "snowflake value should be less than"
                " or equal to 9223372036854775807."
            )

    @classmethod
    def __factory__(cls, string: str) -> Snowflake:
        return cls.from_string(string)

    @classmethod
    def from_string(cls, string: str):
        """Initialize a new Snowflake from a string.

        Parameters
        ----------
        string: :class:`str`
            The snowflake as a string.
        """
        return Snowflake(int(string))

    @property
    def timestamp(self) -> int:
        """:class:`int`: Milliseconds since Discord Epoch,
        the first second of 2015 or 14200704000000
        """
        return self >> 22

    @property
    def worker_id(self) -> int:
        """:class:`int`: Internal worker ID"""
        return (self >> 17) % 16

    @property
    def process_id(self) -> int:
        """:class:`int`: Internal process ID"""
        return (self >> 12) % 16

    @property
    def increment(self) -> int:
        """:class:`int`: For every ID that is generated on that process,
        this number is incremented.
        """
        return self % 2048

    @property
    def unix(self) -> int:
        return self.timestamp + 1420070400000



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import copy
import logging
from dataclasses import dataclass, fields, _is_dataclass_instance
from enum import Enum, EnumMeta
from inspect import getfullargspec
from typing import (
    Dict, Tuple, Union, Generic, TypeVar, Any, TYPE_CHECKING,
    List, get_type_hints, get_origin, get_args
)

from .conversion import convert
from .types import MissingType, MISSING, TypeCache
from ..exceptions import InvalidArgumentAnnotation

if TYPE_CHECKING:
    from ..client import Client
    from ..core.http import HTTPClient

T = TypeVar("T")

_log = logging.getLogger(__package__)


def _asdict_ignore_none(obj: Generic[T]) -> Union[Tuple, Dict, T]:
    """
    Returns a dict from a dataclass that ignores
    all values that are None
    Modification of _asdict_inner from dataclasses

    :param obj:
        Dataclass obj
    """

    if _is_dataclass_instance(obj):
        result = []
        for f in fields(obj):
            value = _asdict_ignore_none(getattr(obj, f.name))

            if isinstance(value, Enum):
                result.append((f.name, value.value))
            # This if statement was added to the function
            elif not isinstance(value, MissingType):
                result.append((f.name, value))

        return dict(result)

    elif isinstance(obj, tuple) and hasattr(obj, '_fields'):
        return type(obj)(*[_asdict_ignore_none(v) for v in obj])

    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_ignore_none(v) for v in obj)

    elif isinstance(obj, dict):
        return type(obj)(
            (
                _asdict_ignore_none(k),
                _asdict_ignore_none(v)
            ) for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


class HTTPMeta(type):
    __meta_items: List[str] = ["_client", "_http"]
    __ori_annotations: Dict[str, type] = {}

    def __new__(mcs, name, base, mapping):
        # Iterates through the meta items, these are keys whom should
        # be added to every object. But to keep typehints we have to
        # define those in the parent class. Yet this gives a conflict
        # because the value is not defined. (that's why we remove it)
        for key in HTTPMeta.__meta_items:
            if mapping.get("__annotations__") and \
                    (value := mapping["__annotations__"].get(key)):
                # We want to keep the type annotations of the objects
                # tho, so lets statically store them so we can read
                # them later.
                HTTPMeta.__ori_annotations.update({key: value})
                del mapping["__annotations__"][key]

        # Instantiate our object
        http_object = super().__new__(mcs, name, base, mapping)

        # Read all removed items
        if getattr(http_object, "__annotations__", None):
            for k, v in HTTPMeta.__ori_annotations.items():
                http_object.__annotations__[k] = v
                setattr(http_object, k, None)

        return http_object


@dataclass
class APIObject(metaclass=HTTPMeta):
    """
    Represents an object which has been fetched from the Discord API.
    """
    _client: Client
    _http: HTTPClient

    def __get_types(self, attr: str, arg_type: type) -> Tuple[type]:
        """Get the types from type annotations.

        Parameters
        ----------
        attr: :class:`str`
            The attribute the typehint belongs to.
        arg_type: :class:`type`
            The type annotation for the attribute.

        Returns
        -------
        Tuple[:class:`type`]
            A collection of type annotation(s). Will most of the times
            consist of 1 item.

        Raises
        ------
        :class:`~pincer.exceptions.InvalidArgumentAnnotation`
            Exception which is raised when the type annotation has not enough
            or too many arguments for the parser to handle.
        """
        origin = get_origin(arg_type)

        if origin is Union:
            # Ahh yes, typing module has no type annotations for this...
            # noinspection PyTypeChecker
            args: Tuple[type] = get_args(arg_type)

            if 2 <= len(args) < 4:
                return args

            raise InvalidArgumentAnnotation(
                f"Attribute `{attr}` in `{type(self).__name__}` has too many "
                f"or not enough arguments! (got {len(args)} expected 2-3)"
            )

        return arg_type,

    def __attr_convert(self, attr: str, attr_type: T) -> T:
        """Convert an attribute to the requested attribute type using
        the factory or the __init__.

        Parameters
        ----------
        attr: :class:`str`
            The attribute the typehint belongs to.
        arg_type: T
            The type annotation for the attribute.

        Returns
        -------
        T
            The instantiated version of the arg_type.
        """
        factory = attr_type

        # Always use `__factory__` over __init__
        if getattr(attr_type, "__factory__", None):
            factory = attr_type.__factory__

        return convert(
            getattr(self, attr),
            factory,
            attr_type,
            self._client
        )

    def __post_init__(self):
        TypeCache()

        # Get all type annotations for the attributes.
        attributes = get_type_hints(self, globalns=TypeCache.cache).items()

        for attr, attr_type in attributes:
            # Ignore private attributes.
            if attr.startswith('_'):
                continue

            types = self.__get_types(attr, attr_type)

            types = tuple(filter(
                lambda tpe: tpe is not None and tpe is not MISSING,
                types
            ))

            if not types:
                raise InvalidArgumentAnnotation(
                    f"Attribute `{attr}` in `{type(self).__name__}` only "
                    "consisted of missing/optional type!"
                )

            specific_tp = types[0]

            if tp := get_origin(specific_tp):
                specific_tp = tp

            if isinstance(specific_tp, EnumMeta) and not getattr(self, attr):
                attr_value = MISSING
            else:
                attr_value = self.__attr_convert(attr, specific_tp)

            setattr(self, attr, attr_value)

    # Set default factory method to from_dict for APIObject's.
    @classmethod
    def __factory__(cls: Generic[T], *args, **kwargs) -> T:
        return cls.from_dict(*args, **kwargs)

    def __str__(self):
        if self.__dict__.get('name'):
            return self.name

        return super().__str__()

    @classmethod
    def from_dict(
            cls: Generic[T],
            data: Dict[str, Union[str, bool, int, Any]]
    ) -> T:
        """
        Parse an API object from a dictionary.
        """
        if isinstance(data, cls):
            return data

        # Disable inspection for IDE because this is valid code for the
        # inherited classes:
        # noinspection PyArgumentList
        return cls(**dict(map(
            lambda key: (
                key,
                data[key].value if isinstance(data[key], Enum) else data[key]
            ),
            filter(
                lambda object_argument: data.get(object_argument) is not None,
                getfullargspec(cls.__init__).args
            )
        )))

    def to_dict(self) -> Dict:
        """
        Transform the current object to a dictionary representation.
        """
        return _asdict_ignore_none(self)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when a user's voice state changes in a subscribed voice channel
(mute, volume, etc.)
"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.webhook import WebhooksUpdateEvent
from ..utils.conversion import construct_client_dict


async def webhooks_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_webhooks_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the webhooks update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.webhook.WebhooksUpdateEvent`]]
        ``on_webhooks_update`` and a ``WebhooksUpdateEvent``
    """
    return "on_webhooks_update", [
        WebhooksUpdateEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export():
    return webhooks_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List, Optional

    from ..user.user import User
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


@dataclass
class Application(APIObject):
    """Represents a Discord application. (eg Bot, OAuth)

    Attributes
    ----------
    bot_public: :class:`bool`
        when false only app owner can join the app's bot to guilds
    bot_require_code_grant: :class:`bool`
        when true the app's bot will only join upon completion of the
        full oauth2 code grant flow
    description: :class:`str`
        the description of the app
    id: :class:`~pincer.utils.snowflake.Snowflake`
        the id of the app
    icon: Optional[:class:`str`]
        the icon hash of the app
    name: :class:`str`
        the name of the app
    privacy_policy_url: APINullable[:class:`str`]
        the url of the app's privacy policy
    summary: :class:`str`
        if this application is a game sold on Discord, this field will be the
        summary field for the store page of its primary sku
    verify_key: :class:`str`
        the hex encoded key for verification in interactions and the GameSDK's
        GetTicket
    cover_image: APINullable[:class:`str`]
        the application's default rich presence invite cover image hash
    flags: APINullable[:class:`int`]
        the application's public flags
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        if this application is a game sold on Discord, this field will be the
        guild to which it has been linked
    owner: APINullable[:class:`~pincer.objects.user.user.User`]
        partial user object containing info on the owner of the application
    primary_sku_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        if this application is a game sold on Discord, this field will be the
        id of the "Game SKU" that is created, if exists
    rpc_origins: APINullable[List[:class:`str`]]
        an array of rpc origin urls, if rpc is enabled
    slug: APINullable[:class:`str`]
        if this application is a game sold on Discord, this field will be the
        URL slug that links to the store page
    terms_of_service_url: APINullable[:class:`str`]
        the url of the app's terms of service
    """
    bot_public: bool
    bot_require_code_grant: bool
    description: str
    id: Snowflake
    icon: Optional[str]
    name: str
    privacy_policy_url: APINullable[str]
    summary: str
    verify_key: str

    cover_image: APINullable[str] = MISSING
    flags: APINullable[int] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    owner: APINullable[User] = MISSING
    primary_sku_id: APINullable[Snowflake] = MISSING
    rpc_origins: APINullable[List[str]] = MISSING
    slug: APINullable[str] = MISSING
    terms_of_service_url: APINullable[str] = MISSING
